from contextlib import asynccontextmanager
import json
import click
from click.exceptions import ClickException

from planet.cli.io import echo_json
from planet.clients.features import FeaturesClient
from planet.geojson import split_ref

from .cmds import command
from .options import compact, limit
from .session import CliSession


@asynccontextmanager
async def features_client(ctx):
    async with CliSession() as sess:
        cl = FeaturesClient(sess, base_url=ctx.obj['BASE_URL'])
        yield cl


@click.group()  # type: ignore
@click.pass_context
@click.option('-u',
              '--base-url',
              default=None,
              help='Assign custom base Features API URL.')
def features(ctx, base_url):
    """Commands for interacting with the Features API"""
    ctx.obj['BASE_URL'] = base_url


@features.group()  # type: ignore
def collections():
    """Commands for interacting with Features API collections"""
    pass


@command(collections, name="create")
@click.option("-t",
              "--title",
              required=True,
              help="a title for the collection")
@click.option("-d",
              "--description",
              "--desc",
              required=False,
              help="a description for the collection")
async def collection_create(ctx, title, description, pretty):
    """Create a new Features API collection.

    Example:

    \b
    planet features collections create \\
      --title "new collection" \\
      --desc "my new collection"
    """
    async with features_client(ctx) as cl:
        col = await cl.create_collection(title, description)
        echo_json(col, pretty)


@command(collections, name="list", extra_args=[limit, compact])
async def collections_list(ctx, pretty, limit, compact):
    """List Features API collections

    Example:

    planet features collections list
    """
    async with features_client(ctx) as cl:
        results = cl.list_collections(limit=limit)

        if compact:
            compact_fields = ('id', 'title', 'description')
            output = [{
                k: v
                for k, v in row.items() if k in compact_fields
            } async for row in results]
        else:
            output = [c async for c in results]

        echo_json(output, pretty)


@command(collections, name="get")
@click.argument("collection_id", required=True)
async def collection_get(ctx, collection_id, pretty):
    """Get a collection by ID

    Example:

    planet features collections get
    """
    async with features_client(ctx) as cl:
        result = await cl.get_collection(collection_id)
        echo_json(result, pretty)


@features.group()
def items():
    """commands for interacting with Features API items (features
    within a collection)"""
    pass


@command(items, name="list", extra_args=[limit])
@click.argument("collection_id", required=True)
async def items_list(ctx, collection_id, pretty, limit):
    """List features in a Features API collection

    Example:

    planet features items list my-collection-123
    """
    async with features_client(ctx) as cl:
        results = cl.list_items(collection_id, limit=limit)
        echo_json([f async for f in results], pretty)


@command(items, name="get")
@click.argument("collection_id")
@click.argument("feature_id", required=False)
async def item_get(ctx, collection_id, feature_id, pretty):
    """Get a feature in a collection.

    You may supply either a collection ID and a feature ID, or
    a feature reference.

    Example:

    planet features items get my-collection-123 item123
    planet features items get "pl:features/my/my-collection-123/item123"
    """

    # ensure that either collection_id and feature_id were supplied, or that
    # a feature ref was supplied as a single value.
    if not ((collection_id and feature_id) or
            ("pl:features" in collection_id)):
        raise ClickException(
            "Must supply either collection_id and feature_id, or a valid feature reference."
        )

    if collection_id.startswith("pl:features"):
        collection_id, feature_id = split_ref(collection_id)

    async with features_client(ctx) as cl:
        feature = await cl.get_item(collection_id, feature_id)
        echo_json(feature, pretty)


@command(items, name="add")
@click.argument("collection_id", required=True)
@click.argument("filename", required=True)
async def item_add(ctx, collection_id, filename, pretty):
    """Add features from a geojson file to a collection

    Example:

    planet features items add my-collection-123 ./my_geom.geojson
    """
    async with features_client(ctx) as cl:
        with open(filename) as data:
            try:
                res = await cl.add_items(collection_id, json.load(data))
            except json.decoder.JSONDecodeError:
                raise ClickException(
                    "Only JSON (.json, .geojson) files are supported in the CLI. Please use https://planet.com/features to upload other files."
                )

    echo_json(res, pretty)



# Copyright 2020 Planet Labs, Inc.
# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test example scripts in this directory

This is not an example script. This is a pytest module that provides automated
testing of the example scripts in this directory to ensure they are always
up-to-date and run successfully.

YAY automated testing and up-to-date examples!
"""
import logging
from pathlib import Path
import os
import subprocess
import sys

import pytest

LOGGER = logging.getLogger(__name__)

# All python files in the current directory except this one
# ref: https://stackoverflow.com/a/56813896
SCRIPTS = [
    s for s in Path(__file__).parent.resolve().glob('*.py')
    if s.name != Path(__file__).name
]


# use the script name in the test name
def idfn(script_path):
    return script_path.name


#  provide an environment variable that points to the test temporary download
# directory, to be used by the script for all downloads
@pytest.mark.parametrize('script', SCRIPTS, ids=idfn)
def test_example_script_execution(script, tmpdir):
    completed = subprocess.run(
        [sys.executable, str(script)],
        env={
            'TEST_DOWNLOAD_DIR': str(tmpdir),
            'PL_API_KEY': os.getenv('PL_API_KEY')
        },
        stderr=subprocess.PIPE  # capture stdout for reporting
    )

    assert not completed.returncode, (
        'script failed with following std_err output:\n'
        f'{str(completed.stderr, "utf-8")}')



# Copyright 2022 Planet Labs, PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Functionality for processing inputs and outputs."""
from datetime import datetime
import logging
import typing

from . import exceptions, geojson

LOGGER = logging.getLogger(__name__)

RFC3339_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"


async def collect(
    values: typing.AsyncIterator[dict]
) -> typing.Union[typing.List[dict], dict]:
    """Collect a sequence into JSON.

    If the items in values are GeoJSON features, the JSON blob is a GeoJSON
    FeatureCollection. Otherwise, the JSON blob is a list of the items.

    Returns:
        JSON blob containing all sequence entries.
    """
    as_list = [v async for v in values]

    ret: typing.Union[typing.List[dict], dict]
    try:
        ret = geojson.as_featurecollection(as_list)
    except exceptions.GeoJSONError:
        ret = as_list

    return ret


def str_to_datetime(string: str) -> datetime:
    """Convert a string to a datetime.

    First, tries parsing the string as a RFC 3339 format as returned from the
    Planet APIs (e.g. '2021-01-01T01:40:07.359Z'), then tries parsing
    as an IS0 8601 format using `datetime.fromisoformat()`.
    """
    try:
        dt = datetime.strptime(string, RFC3339_FORMAT)
    except ValueError:
        LOGGER.debug(f'{string} does not match format {RFC3339_FORMAT}')
        try:
            dt = datetime.fromisoformat(string)
        except ValueError:
            raise exceptions.PlanetError(
                f'{string} does not match RFC 3339 or ISO 8601 formats')
    return dt



# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
from contextlib import nullcontext as does_not_raise
import copy
from http import HTTPStatus
import hashlib
import json
import logging
import math
from pathlib import Path

import httpx
import pytest
import respx

from planet import exceptions, DataClient, data_filter
from planet.clients.data import (LIST_SORT_DEFAULT,
                                 LIST_SEARCH_TYPE_DEFAULT,
                                 SEARCH_SORT_DEFAULT)
from planet.sync.data import DataAPI
from planet.http import Session

TEST_URL = 'http://www.mocknotrealurl.com/api/path'
TEST_SEARCHES_URL = f'{TEST_URL}/searches'
TEST_STATS_URL = f'{TEST_URL}/stats'

VALID_SEARCH_ID = '286469f0b27c476e96c3c4e561f59664'

LOGGER = logging.getLogger(__name__)


@pytest.fixture
def search_filter(get_test_file_json):
    filename = 'data_search_filter_2022-01.json'
    return get_test_file_json(filename)


@pytest.fixture
def item_descriptions(get_test_file_json):
    item_ids = [
        '20220125_075509_67_1061',
        '20220125_075511_17_1061',
        '20220125_075650_17_1061'
    ]
    items = [get_test_file_json(f'data_item_{id}.json') for id in item_ids]
    return items


@pytest.fixture
def search_response(item_descriptions):
    response = {
        "type": "FeatureCollection",
        "_links": {
            "_first": "firstpageurl",
            "_next": "nextpageurl",
            "_self": "selfurl"
        },
        "features": item_descriptions
    }

    return response


@pytest.fixture(scope="module")
def data_api():
    return DataAPI(Session(), TEST_URL)


@respx.mock
@pytest.mark.anyio
async def test_search_basic(item_descriptions,
                            search_response,
                            mock_bundles,
                            session):

    quick_search_url = f'{TEST_URL}/quick-search'
    next_page_url = f'{TEST_URL}/blob/?page_marker=IAmATest'

    item1, item2, item3 = item_descriptions
    page1_response = {
        "_links": {
            "_next": next_page_url
        }, "features": [item1, item2]
    }
    mock_resp1 = httpx.Response(HTTPStatus.OK, json=page1_response)
    respx.post(quick_search_url).return_value = mock_resp1

    page2_response = {"_links": {"_self": next_page_url}, "features": [item3]}
    mock_resp2 = httpx.Response(HTTPStatus.OK, json=page2_response)
    respx.get(next_page_url).return_value = mock_resp2

    cl = DataClient(session, base_url=TEST_URL)
    items_list = [i async for i in cl.search(['PSScene'])]

    # check that request is correct
    expected_request = {
        "item_types": ["PSScene"], "filter": data_filter.empty_filter()
    }

    actual_body = json.loads(respx.calls[1].request.content)
    assert actual_body == expected_request

    # check that all of the items were returned unchanged
    assert items_list == item_descriptions


@respx.mock
def test_search_basic_sync(item_descriptions,
                           search_response,
                           mock_bundles,
                           data_api):

    quick_search_url = f'{TEST_URL}/quick-search'
    next_page_url = f'{TEST_URL}/blob/?page_marker=IAmATest'

    item1, item2, item3 = item_descriptions
    page1_response = {
        "_links": {
            "_next": next_page_url
        }, "features": [item1, item2]
    }
    mock_resp1 = httpx.Response(HTTPStatus.OK, json=page1_response)
    respx.post(quick_search_url).return_value = mock_resp1

    page2_response = {"_links": {"_self": next_page_url}, "features": [item3]}
    mock_resp2 = httpx.Response(HTTPStatus.OK, json=page2_response)
    respx.get(next_page_url).return_value = mock_resp2

    items_list = list(data_api.search(['PSScene']))

    # check that request is correct
    expected_request = {
        "item_types": ["PSScene"], "filter": data_filter.empty_filter()
    }
    actual_body = json.loads(respx.calls[0].request.content)
    assert actual_body == expected_request

    # check that all of the items were returned unchanged
    assert items_list == item_descriptions


@respx.mock
@pytest.mark.anyio
async def test_search_name(item_descriptions,
                           search_response,
                           mock_bundles,
                           session):

    quick_search_url = f'{TEST_URL}/quick-search'
    next_page_url = f'{TEST_URL}/blob/?page_marker=IAmATest'

    item1, item2, item3 = item_descriptions
    page1_response = {
        "_links": {
            "_next": next_page_url
        }, "features": [item1, item2]
    }
    mock_resp1 = httpx.Response(HTTPStatus.OK, json=page1_response)
    respx.post(quick_search_url).return_value = mock_resp1

    page2_response = {"_links": {"_self": next_page_url}, "features": [item3]}
    mock_resp2 = httpx.Response(HTTPStatus.OK, json=page2_response)
    respx.get(next_page_url).return_value = mock_resp2

    cl = DataClient(session, base_url=TEST_URL)
    items_list = [i async for i in cl.search(['PSScene'], name='quick_search')]

    # check that request is correct
    expected_request = {
        "item_types": ["PSScene"],
        "filter": data_filter.empty_filter(),
        "name": "quick_search"
    }
    actual_body = json.loads(respx.calls[0].request.content)
    assert actual_body == expected_request

    # check that all of the items were returned unchanged
    assert items_list == item_descriptions


@respx.mock
@pytest.mark.anyio
@pytest.mark.parametrize("geom_fixture", [('geom_geojson'),
                                          ('geom_reference')])
async def test_search_geometry(geom_fixture,
                               item_descriptions,
                               mock_bundles,
                               session,
                               request):

    quick_search_url = f'{TEST_URL}/quick-search'
    next_page_url = f'{TEST_URL}/blob/?page_marker=IAmATest'

    item1, item2, item3 = item_descriptions
    page1_response = {
        "_links": {
            "_next": next_page_url
        }, "features": [item1, item2]
    }
    mock_resp1 = httpx.Response(HTTPStatus.OK, json=page1_response)
    respx.post(quick_search_url).return_value = mock_resp1

    page2_response = {"_links": {"_self": next_page_url}, "features": [item3]}
    mock_resp2 = httpx.Response(HTTPStatus.OK, json=page2_response)
    respx.get(next_page_url).return_value = mock_resp2

    cl = DataClient(session, base_url=TEST_URL)
    geom = request.getfixturevalue(geom_fixture)
    items_list = [
        i async for i in cl.search(['PSScene'], name='quick_search',
                                   geometry=geom)
    ]
    # check that request is correct
    expected_request = {
        "item_types": ["PSScene"],
        "geometry": geom,
        "filter": data_filter.empty_filter(),
        "name": "quick_search"
    }
    actual_body = json.loads(respx.calls[0].request.content)

    assert actual_body == expected_request

    # check that all of the items were returned unchanged
    assert items_list == item_descriptions


@respx.mock
@pytest.mark.parametrize("geom_fixture", [('geom_geojson'),
                                          ('geom_reference')])
def test_search_geometry_sync(geom_fixture,
                              item_descriptions,
                              mock_bundles,
                              data_api,
                              request):

    quick_search_url = f'{TEST_URL}/quick-search'
    next_page_url = f'{TEST_URL}/blob/?page_marker=IAmATest'

    item1, item2, item3 = item_descriptions
    page1_response = {
        "_links": {
            "_next": next_page_url
        }, "features": [item1, item2]
    }
    mock_resp1 = httpx.Response(HTTPStatus.OK, json=page1_response)
    respx.post(quick_search_url).return_value = mock_resp1

    page2_response = {"_links": {"_self": next_page_url}, "features": [item3]}
    mock_resp2 = httpx.Response(HTTPStatus.OK, json=page2_response)
    respx.get(next_page_url).return_value = mock_resp2

    geom = request.getfixturevalue(geom_fixture)
    items_list = list(
        data_api.search(['PSScene'], name='quick_search', geometry=geom))
    # check that request is correct
    expected_request = {
        "item_types": ["PSScene"],
        "geometry": geom,
        "filter": data_filter.empty_filter(),
        "name": "quick_search"
    }
    actual_body = json.loads(respx.calls[0].request.content)

    assert actual_body == expected_request

    # check that all of the items were returned unchanged
    assert items_list == item_descriptions


@respx.mock
@pytest.mark.anyio
async def test_search_filter(item_descriptions,
                             search_filter,
                             search_response,
                             mock_bundles,
                             session):

    quick_search_url = f'{TEST_URL}/quick-search'
    next_page_url = f'{TEST_URL}/blob/?page_marker=IAmATest'

    item1, item2, item3 = item_descriptions
    page1_response = {
        "_links": {
            "_next": next_page_url
        }, "features": [item1, item2]
    }
    mock_resp1 = httpx.Response(HTTPStatus.OK, json=page1_response)
    respx.post(quick_search_url).return_value = mock_resp1

    page2_response = {"_links": {"_self": next_page_url}, "features": [item3]}
    mock_resp2 = httpx.Response(HTTPStatus.OK, json=page2_response)
    respx.get(next_page_url).return_value = mock_resp2

    cl = DataClient(session, base_url=TEST_URL)
    items_list = [
        i async for i in cl.search(item_types=['PSScene'],
                                   search_filter=search_filter)
    ]

    # check that request is correct
    expected_request = {"item_types": ["PSScene"], "filter": search_filter}
    actual_body = json.loads(respx.calls[0].request.content)
    assert actual_body == expected_request

    # check that all of the items were returned unchanged
    assert items_list == item_descriptions


@respx.mock
@pytest.mark.anyio
async def test_search_filter_positional_args(item_descriptions,
                                             search_filter,
                                             search_response,
                                             mock_bundles,
                                             session):
    """test the search method using positional args"""

    quick_search_url = f'{TEST_URL}/quick-search'

    item1, item2, item3 = item_descriptions
    response = {"features": [item1, item2, item3]}
    mock_resp = httpx.Response(HTTPStatus.OK, json=response)
    respx.post(quick_search_url).return_value = mock_resp

    cl = DataClient(session, base_url=TEST_URL)

    # search using a positional arg for the filter.
    items_list = [i async for i in cl.search(['PSScene'], search_filter)]

    # check that request is correct
    expected_request = {"item_types": ["PSScene"], "filter": search_filter}
    actual_body = json.loads(respx.calls[0].request.content)
    assert actual_body == expected_request

    # check that all of the items were returned unchanged
    assert items_list == item_descriptions


@respx.mock
@pytest.mark.anyio
async def test_search_sort(item_descriptions,
                           search_filter,
                           search_response,
                           mock_bundles,
                           session):

    sort = 'acquired asc'
    quick_search_url = f'{TEST_URL}/quick-search?_sort={sort}'

    item1, _, _ = item_descriptions
    page1_response = {"_links": {}, "features": [item1]}
    mock_resp1 = httpx.Response(HTTPStatus.OK, json=page1_response)
    respx.post(quick_search_url).return_value = mock_resp1

    # if the sort parameter is not used correctly, the client will not send
    # the request to the mocked endpoint and this test will fail
    cl = DataClient(session, base_url=TEST_URL)

    # run through the iterator to actually initiate the call
    [
        i async for i in cl.search(['PSScene'], search_filter=search_filter,
                                   sort=sort)
    ]


@respx.mock
@pytest.mark.anyio
async def test_search_limit(item_descriptions,
                            search_filter,
                            search_response,
                            mock_bundles,
                            session):

    quick_search_url = f'{TEST_URL}/quick-search'

    page_response = {
        "_links": {},
        "features": item_descriptions  # three items
    }
    mock_resp = httpx.Response(HTTPStatus.OK, json=page_response)
    respx.post(quick_search_url).return_value = mock_resp

    cl = DataClient(session, base_url=TEST_URL)
    items_list = [
        i async for i in cl.search(['PSScene'], search_filter=search_filter,
                                   limit=2)
    ]

    # check only the first two results were returned
    assert items_list == item_descriptions[:2]


@respx.mock
@pytest.mark.anyio
async def test_create_search_basic(search_filter, mock_bundles, session):

    page_response = {
        "__daily_email_enabled": False,
        "_links": {
            "_self": "string", "thumbnail": "string"
        },
        "created": "2019-08-24T14:15:22Z",
        "filter": search_filter,
        "id": "string",
        "last_executed": "2019-08-24T14:15:22Z",
        "name": "test",
        "updated": "2019-08-24T14:15:22Z"
    }
    mock_resp = httpx.Response(HTTPStatus.OK, json=page_response)
    respx.post(TEST_SEARCHES_URL).return_value = mock_resp

    cl = DataClient(session, base_url=TEST_URL)
    search = await cl.create_search(item_types=['PSScene'],
                                    search_filter=search_filter,
                                    name='test')

    # check that request is correct
    expected_request = {
        "item_types": ["PSScene"],
        "filter": search_filter,
        "name": "test",
        "__daily_email_enabled": False
    }
    actual_body = json.loads(respx.calls[0].request.content)
    assert actual_body == expected_request

    # check the response is returned unaltered
    assert search == page_response


@respx.mock
def test_create_search_basic_sync(search_filter, mock_bundles, data_api):

    page_response = {
        "__daily_email_enabled": False,
        "_links": {
            "_self": "string", "thumbnail": "string"
        },
        "created": "2019-08-24T14:15:22Z",
        "filter": search_filter,
        "id": "string",
        "last_executed": "2019-08-24T14:15:22Z",
        "name": "test",
        "updated": "2019-08-24T14:15:22Z"
    }
    mock_resp = httpx.Response(HTTPStatus.OK, json=page_response)
    respx.post(TEST_SEARCHES_URL).return_value = mock_resp

    search = data_api.create_search(item_types=['PSScene'],
                                    search_filter=search_filter,
                                    name='test')

    # check that request is correct
    expected_request = {
        "item_types": ["PSScene"],
        "filter": search_filter,
        "name": "test",
        "__daily_email_enabled": False
    }
    actual_body = json.loads(respx.calls[0].request.content)
    assert actual_body == expected_request

    # check the response is returned unaltered
    assert search == page_response


@respx.mock
@pytest.mark.anyio
async def test_create_search_basic_positional_args(search_filter,
                                                   mock_bundles,
                                                   session):
    """Test that positional arguments are accepted for create_search"""

    page_response = {
        "__daily_email_enabled": False,
        "_links": {
            "_self": "string", "thumbnail": "string"
        },
        "created": "2019-08-24T14:15:22Z",
        "filter": search_filter,
        "id": "string",
        "last_executed": "2019-08-24T14:15:22Z",
        "name": "test",
        "updated": "2019-08-24T14:15:22Z"
    }
    mock_resp = httpx.Response(HTTPStatus.OK, json=page_response)
    respx.post(TEST_SEARCHES_URL).return_value = mock_resp

    cl = DataClient(session, base_url=TEST_URL)
    search = await cl.create_search(['PSScene'], search_filter, name='test')

    # check that request is correct
    expected_request = {
        "item_types": ["PSScene"],
        "filter": search_filter,
        "name": "test",
        "__daily_email_enabled": False
    }
    actual_body = json.loads(respx.calls[0].request.content)
    assert actual_body == expected_request

    # check the response is returned unaltered
    assert search == page_response


@respx.mock
@pytest.mark.anyio
async def test_create_search_email(search_filter, mock_bundles, session):

    page_response = {
        "__daily_email_enabled": True,
        "_links": {
            "_self": "string", "thumbnail": "string"
        },
        "created": "2019-08-24T14:15:22Z",
        "filter": search_filter,
        "id": "string",
        "last_executed": "2019-08-24T14:15:22Z",
        "name": "test",
        "updated": "2019-08-24T14:15:22Z"
    }
    mock_resp = httpx.Response(HTTPStatus.OK, json=page_response)
    respx.post(TEST_SEARCHES_URL).return_value = mock_resp

    cl = DataClient(session, base_url=TEST_URL)
    search = await cl.create_search(['PSScene'],
                                    search_filter=search_filter,
                                    name='test',
                                    enable_email=True)

    # check that request is correct
    expected_request = {
        "item_types": ["PSScene"],
        "filter": search_filter,
        "name": "test",
        "__daily_email_enabled": True
    }
    actual_body = json.loads(respx.calls[0].request.content)
    assert actual_body == expected_request

    # check the response is returned unaltered
    assert search == page_response


@respx.mock
@pytest.mark.anyio
async def test_get_search_success(search_id, search_result, session):
    get_url = f'{TEST_SEARCHES_URL}/{search_id}'
    mock_resp = httpx.Response(HTTPStatus.OK, json=search_result)
    respx.get(get_url).return_value = mock_resp
    cl = DataClient(session, base_url=TEST_URL)
    search = await cl.get_search(search_id)
    assert search_result == search


@respx.mock
def test_get_search_success_sync(search_id, search_result, data_api):
    get_url = f'{TEST_SEARCHES_URL}/{search_id}'
    mock_resp = httpx.Response(HTTPStatus.OK, json=search_result)
    respx.get(get_url).return_value = mock_resp
    search = data_api.get_search(search_id)
    assert search_result == search


@respx.mock
@pytest.mark.anyio
async def test_get_search_id_doesnt_exist(search_id, session):
    get_url = f'{TEST_SEARCHES_URL}/{search_id}'

    resp = {
        "message": f'The requested search id does not exist:\
        {search_id}'
    }
    mock_resp = httpx.Response(404, json=resp)
    respx.get(get_url).return_value = mock_resp
    cl = DataClient(session, base_url=TEST_URL)

    with pytest.raises(exceptions.MissingResource):
        await cl.get_search(search_id)


@respx.mock
@pytest.mark.anyio
async def test_update_search_basic(search_filter, session):

    page_response = {
        "__daily_email_enabled": False,
        "_links": {
            "_self": "string", "thumbnail": "string"
        },
        "created": "2019-08-24T14:15:22Z",
        "filter": search_filter,
        "id": VALID_SEARCH_ID,
        "last_executed": "2019-08-24T14:15:22Z",
        "name": "test",
        "updated": "2019-08-24T14:15:22Z"
    }
    mock_resp = httpx.Response(HTTPStatus.OK, json=page_response)
    respx.put(
        f'{TEST_SEARCHES_URL}/{VALID_SEARCH_ID}').return_value = mock_resp

    cl = DataClient(session, base_url=TEST_URL)
    search = await cl.update_search(VALID_SEARCH_ID,
                                    item_types=['PSScene'],
                                    search_filter=search_filter,
                                    name='test')

    # check that request is correct
    expected_request = {
        "item_types": ["PSScene"],
        "filter": search_filter,
        "name": "test",
        "__daily_email_enabled": False
    }
    actual_body = json.loads(respx.calls[0].request.content)
    assert actual_body == expected_request

    # check the response is returned unaltered
    assert search == page_response


@respx.mock
def test_update_search_basic_sync(search_filter, data_api):

    page_response = {
        "__daily_email_enabled": False,
        "_links": {
            "_self": "string", "thumbnail": "string"
        },
        "created": "2019-08-24T14:15:22Z",
        "filter": search_filter,
        "id": VALID_SEARCH_ID,
        "last_executed": "2019-08-24T14:15:22Z",
        "name": "test",
        "updated": "2019-08-24T14:15:22Z"
    }
    mock_resp = httpx.Response(HTTPStatus.OK, json=page_response)
    respx.put(
        f'{TEST_SEARCHES_URL}/{VALID_SEARCH_ID}').return_value = mock_resp

    search = data_api.update_search(VALID_SEARCH_ID,
                                    item_types=['PSScene'],
                                    search_filter=search_filter,
                                    name='test')

    # check that request is correct
    expected_request = {
        "item_types": ["PSScene"],
        "filter": search_filter,
        "name": "test",
        "__daily_email_enabled": False
    }
    actual_body = json.loads(respx.calls[0].request.content)
    assert actual_body == expected_request

    # check the response is returned unaltered
    assert search == page_response


@respx.mock
@pytest.mark.anyio
async def test_update_search_basic_positional_args(search_filter, session):
    """Test that positional arguments are accepted for update_search"""

    page_response = {
        "__daily_email_enabled": False,
        "_links": {
            "_self": "string", "thumbnail": "string"
        },
        "created": "2019-08-24T14:15:22Z",
        "filter": search_filter,
        "id": VALID_SEARCH_ID,
        "last_executed": "2019-08-24T14:15:22Z",
        "name": "test",
        "updated": "2019-08-24T14:15:22Z"
    }
    mock_resp = httpx.Response(HTTPStatus.OK, json=page_response)
    respx.put(
        f'{TEST_SEARCHES_URL}/{VALID_SEARCH_ID}').return_value = mock_resp

    cl = DataClient(session, base_url=TEST_URL)
    search = await cl.update_search(VALID_SEARCH_ID, ['PSScene'],
                                    search_filter,
                                    name='test')

    # check that request is correct
    expected_request = {
        "item_types": ["PSScene"],
        "filter": search_filter,
        "name": "test",
        "__daily_email_enabled": False
    }
    actual_body = json.loads(respx.calls[0].request.content)
    assert actual_body == expected_request

    # check the response is returned unaltered
    assert search == page_response


@respx.mock
@pytest.mark.anyio
@pytest.mark.parametrize("limit, expected_list_length", [(None, 4), (3, 3)])
async def test_list_searches_success(limit,
                                     expected_list_length,
                                     search_result,
                                     session):
    page1_response = {"_links": {}, "searches": [search_result] * 4}
    route = respx.get(TEST_SEARCHES_URL)
    route.return_value = httpx.Response(200, json=page1_response)

    cl = DataClient(session, base_url=TEST_URL)

    assert len([s async for s in cl.list_searches(limit=limit)
                ]) == expected_list_length

    assert route.called


@respx.mock
@pytest.mark.parametrize("limit, expected_list_length", [(None, 4), (3, 3)])
def test_list_searches_success_sync(limit,
                                    expected_list_length,
                                    search_result,
                                    data_api):
    page1_response = {"_links": {}, "searches": [search_result] * 4}
    route = respx.get(TEST_SEARCHES_URL)
    route.return_value = httpx.Response(200, json=page1_response)

    assert len(list(
        data_api.list_searches(limit=limit))) == expected_list_length

    assert route.called


@respx.mock
@pytest.mark.anyio
@pytest.mark.parametrize("sort, rel_url",
                         [(LIST_SORT_DEFAULT, ''),
                          ('created asc', '?_sort=created+asc')])
async def test_list_searches_sort(sort, rel_url, search_result, session):
    page1_response = {"_links": {}, "searches": [search_result] * 4}
    route = respx.get(f'{TEST_SEARCHES_URL}{rel_url}')
    route.return_value = httpx.Response(200, json=page1_response)

    cl = DataClient(session, base_url=TEST_URL)
    _ = [s async for s in cl.list_searches(sort=sort)]

    assert route.called


@respx.mock
@pytest.mark.anyio
@pytest.mark.parametrize("search_type, rel_url",
                         [(LIST_SEARCH_TYPE_DEFAULT, ''),
                          ('saved', '?search_type=saved')])
async def test_list_searches_searchtype(search_type,
                                        rel_url,
                                        search_result,
                                        session):
    page1_response = {"_links": {}, "searches": [search_result] * 4}
    route = respx.get(f'{TEST_SEARCHES_URL}{rel_url}')
    route.return_value = httpx.Response(200, json=page1_response)

    cl = DataClient(session, base_url=TEST_URL)
    _ = [s async for s in cl.list_searches(search_type=search_type)]

    assert route.called


@respx.mock
@pytest.mark.anyio
@pytest.mark.parametrize(
    "sort, search_type, expectation",
    [('DOESNOTEXIST', 'ANY', pytest.raises(exceptions.ClientError)),
     ('CREATED DESC', 'DOESNOTEXIST', pytest.raises(exceptions.ClientError))])
async def test_list_searches_args_do_not_match(sort,
                                               search_type,
                                               expectation,
                                               session):
    route = respx.get(TEST_SEARCHES_URL)
    route.return_value = httpx.Response(200, json={})

    cl = DataClient(session, base_url=TEST_URL)

    with expectation:
        [s async for s in cl.list_searches(sort=sort, search_type=search_type)]

    assert not route.called


@respx.mock
@pytest.mark.anyio
@pytest.mark.parametrize("retcode, expectation",
                         [(204, does_not_raise()),
                          (404, pytest.raises(exceptions.APIError))])
async def test_delete_search(retcode, expectation, session):
    mock_resp = httpx.Response(retcode)
    route = respx.delete(f'{TEST_SEARCHES_URL}/{VALID_SEARCH_ID}')
    route.return_value = mock_resp
    cl = DataClient(session, base_url=TEST_URL)

    with expectation:
        await cl.delete_search(VALID_SEARCH_ID)

    assert route.called


@respx.mock
@pytest.mark.parametrize("retcode, expectation",
                         [(204, does_not_raise()),
                          (404, pytest.raises(exceptions.APIError))])
def test_delete_search_sync(retcode, expectation, data_api):
    mock_resp = httpx.Response(retcode)
    route = respx.delete(f'{TEST_SEARCHES_URL}/{VALID_SEARCH_ID}')
    route.return_value = mock_resp

    with expectation:
        data_api.delete_search(VALID_SEARCH_ID)

    assert route.called


@respx.mock
@pytest.mark.anyio
@pytest.mark.parametrize("search_id, valid", [(VALID_SEARCH_ID, True),
                                              ('invalid', False)])
@pytest.mark.parametrize("limit, expected_count", [(None, 3), (2, 2)])
async def test_run_search_basic(item_descriptions,
                                session,
                                search_id,
                                valid,
                                limit,
                                expected_count):
    """Ensure run_search is successful and handles search_id and limit"""
    next_page_url = f'{TEST_URL}/blob/?page_marker=IAmATest'
    item1, item2, item3 = item_descriptions
    page1_response = {
        "_links": {
            "_next": next_page_url
        }, "features": [item1, item2]
    }

    route = respx.get(f'{TEST_SEARCHES_URL}/{search_id}/results')
    route.return_value = httpx.Response(204, json=page1_response)

    page2_response = {"_links": {"_self": next_page_url}, "features": [item3]}
    mock_resp2 = httpx.Response(HTTPStatus.OK, json=page2_response)
    respx.get(next_page_url).return_value = mock_resp2

    cl = DataClient(session, base_url=TEST_URL)

    if valid:
        items_list = [i async for i in cl.run_search(search_id, limit=limit)]

        assert route.called

        # check that all of the items were returned unchanged
        assert items_list == item_descriptions[:expected_count]
    else:
        with pytest.raises(exceptions.ClientError):
            [i async for i in cl.run_search(search_id)]


@respx.mock
@pytest.mark.parametrize("search_id, valid", [(VALID_SEARCH_ID, True),
                                              ('invalid', False)])
@pytest.mark.parametrize("limit, expected_count", [(None, 3), (2, 2)])
def test_run_search_basic_sync(item_descriptions,
                               data_api,
                               search_id,
                               valid,
                               limit,
                               expected_count):
    """Ensure run_search is successful and handles search_id and limit"""
    next_page_url = f'{TEST_URL}/blob/?page_marker=IAmATest'
    item1, item2, item3 = item_descriptions
    page1_response = {
        "_links": {
            "_next": next_page_url
        }, "features": [item1, item2]
    }

    route = respx.get(f'{TEST_SEARCHES_URL}/{search_id}/results')
    route.return_value = httpx.Response(204, json=page1_response)

    page2_response = {"_links": {"_self": next_page_url}, "features": [item3]}
    mock_resp2 = httpx.Response(HTTPStatus.OK, json=page2_response)
    respx.get(next_page_url).return_value = mock_resp2

    if valid:
        items_list = list(data_api.run_search(search_id, limit=limit))

        assert route.called

        # check that all of the items were returned unchanged
        assert items_list == item_descriptions[:expected_count]
    else:
        with pytest.raises(exceptions.ClientError):
            list(data_api.run_search(search_id))


@respx.mock
@pytest.mark.anyio
@pytest.mark.parametrize("sort, rel_url, valid",
                         [(SEARCH_SORT_DEFAULT, '', True),
                          ('acquired asc', '?_sort=acquired+asc', True),
                          ('invalid', '', False)])
async def test_run_search_sort(item_descriptions,
                               session,
                               sort,
                               rel_url,
                               valid):
    next_page_url = f'{TEST_URL}/blob/?page_marker=IAmATest'
    item1, item2, item3 = item_descriptions
    page1_response = {
        "_links": {
            "_next": next_page_url
        }, "features": [item1, item2]
    }

    route = respx.get(
        f'{TEST_SEARCHES_URL}/{VALID_SEARCH_ID}/results{rel_url}')
    route.return_value = httpx.Response(204, json=page1_response)

    page2_response = {"_links": {"_self": next_page_url}, "features": [item3]}
    mock_resp2 = httpx.Response(HTTPStatus.OK, json=page2_response)
    respx.get(next_page_url).return_value = mock_resp2

    cl = DataClient(session, base_url=TEST_URL)

    expectation = pytest.raises(
        exceptions.ClientError) if not valid else does_not_raise()

    with expectation:
        [s async for s in cl.run_search(VALID_SEARCH_ID, sort=sort)]
        assert route.called


@respx.mock
@pytest.mark.anyio
async def test_run_search_doesnotexist(session):
    route = respx.get(f'{TEST_SEARCHES_URL}/{VALID_SEARCH_ID}/results')
    route.return_value = httpx.Response(404)

    cl = DataClient(session, base_url=TEST_URL)
    with pytest.raises(exceptions.APIError):
        [i async for i in cl.run_search(VALID_SEARCH_ID)]

    assert route.called


@respx.mock
@pytest.mark.anyio
async def test_get_stats_success(search_filter, session):

    page_response = {
        "buckets": [
            {
                "count": 433638, "start_time": "2022-01-01T00:00:00.000000Z"
            },
            {
                "count": 431924, "start_time": "2022-01-02T00:00:00.000000Z"
            },
            {
                "count": 417138, "start_time": "2022-01-03T00:00:00.000000Z"
            },
        ],
    }
    mock_resp = httpx.Response(HTTPStatus.OK, json=page_response)
    respx.post(TEST_STATS_URL).return_value = mock_resp

    cl = DataClient(session, base_url=TEST_URL)
    stats = await cl.get_stats(['PSScene'], search_filter, 'day')

    # check that request is correct
    expected_request = {
        "item_types": ["PSScene"], "filter": search_filter, "interval": "day"
    }
    actual_body = json.loads(respx.calls[0].request.content)
    assert actual_body == expected_request

    # check the response is returned unaltered
    assert stats == page_response


@respx.mock
def test_get_stats_success_sync(search_filter, data_api):

    page_response = {
        "buckets": [
            {
                "count": 433638, "start_time": "2022-01-01T00:00:00.000000Z"
            },
            {
                "count": 431924, "start_time": "2022-01-02T00:00:00.000000Z"
            },
            {
                "count": 417138, "start_time": "2022-01-03T00:00:00.000000Z"
            },
        ],
    }
    mock_resp = httpx.Response(HTTPStatus.OK, json=page_response)
    respx.post(TEST_STATS_URL).return_value = mock_resp

    stats = data_api.get_stats(['PSScene'], search_filter, 'day')

    # check that request is correct
    expected_request = {
        "item_types": ["PSScene"], "filter": search_filter, "interval": "day"
    }
    actual_body = json.loads(respx.calls[0].request.content)
    assert actual_body == expected_request

    # check the response is returned unaltered
    assert stats == page_response


@respx.mock
@pytest.mark.anyio
async def test_get_stats_invalid_interval(search_filter, session):
    cl = DataClient(session, base_url=TEST_URL)

    with pytest.raises(exceptions.ClientError):
        await cl.get_stats(['PSScene'], search_filter, 'invalid')


@respx.mock
@pytest.mark.anyio
async def test_list_item_assets_success(session):
    item_type_id = 'PSScene'
    item_id = '20221003_002705_38_2461'
    assets_url = f'{TEST_URL}/item-types/{item_type_id}/items/{item_id}/assets'

    page_response = {
        "basic_analytic_4b": {
            "_links": {
                "_self": "SELFURL",
                "activate": "ACTIVATEURL",
                "type": "https://api.planet.com/data/v1/asset-types/basic_analytic_4b"
            },
            "_permissions": ["download"],
            "md5_digest": None,
            "status": "inactive",
            "type": "basic_analytic_4b"
        },
        "basic_udm2": {
            "_links": {
                "_self": "SELFURL",
                "activate": "ACTIVATEURL",
                "type": "https://api.planet.com/data/v1/asset-types/basic_udm2"
            },
            "_permissions": ["download"],
            "md5_digest": None,
            "status": "inactive",
            "type": "basic_udm2"
        }
    }
    mock_resp = httpx.Response(HTTPStatus.OK, json=page_response)
    respx.get(assets_url).return_value = mock_resp

    cl = DataClient(session, base_url=TEST_URL)
    assets = await cl.list_item_assets(item_type_id, item_id)

    # check the response is returned unaltered
    assert assets == page_response


@respx.mock
def test_list_item_assets_success_sync(data_api):
    item_type_id = 'PSScene'
    item_id = '20221003_002705_38_2461'
    assets_url = f'{TEST_URL}/item-types/{item_type_id}/items/{item_id}/assets'

    page_response = {
        "basic_analytic_4b": {
            "_links": {
                "_self": "SELFURL",
                "activate": "ACTIVATEURL",
                "type": "https://api.planet.com/data/v1/asset-types/basic_analytic_4b"
            },
            "_permissions": ["download"],
            "md5_digest": None,
            "status": "inactive",
            "type": "basic_analytic_4b"
        },
        "basic_udm2": {
            "_links": {
                "_self": "SELFURL",
                "activate": "ACTIVATEURL",
                "type": "https://api.planet.com/data/v1/asset-types/basic_udm2"
            },
            "_permissions": ["download"],
            "md5_digest": None,
            "status": "inactive",
            "type": "basic_udm2"
        }
    }
    mock_resp = httpx.Response(HTTPStatus.OK, json=page_response)
    respx.get(assets_url).return_value = mock_resp

    assets = data_api.list_item_assets(item_type_id, item_id)

    # check the response is returned unaltered
    assert assets == page_response


@respx.mock
@pytest.mark.anyio
async def test_list_item_assets_missing(session):
    item_type_id = 'PSScene'
    item_id = '20221003_002705_38_2461xx'
    assets_url = f'{TEST_URL}/item-types/{item_type_id}/items/{item_id}/assets'

    mock_resp = httpx.Response(404)
    respx.get(assets_url).return_value = mock_resp

    cl = DataClient(session, base_url=TEST_URL)

    with pytest.raises(exceptions.APIError):
        await cl.list_item_assets(item_type_id, item_id)


@respx.mock
@pytest.mark.anyio
@pytest.mark.parametrize("asset_type_id, expectation",
                         [('basic_udm2', does_not_raise()),
                          ('invalid', pytest.raises(exceptions.ClientError))])
async def test_get_asset(asset_type_id, expectation, mock_bundles, session):
    item_type_id = 'PSScene'
    item_id = '20221003_002705_38_2461'
    assets_url = f'{TEST_URL}/item-types/{item_type_id}/items/{item_id}/assets'

    basic_udm2_asset = {
        "_links": {
            "_self": "SELFURL",
            "activate": "ACTIVATEURL",
            "type": "https://api.planet.com/data/v1/asset-types/basic_udm2"
        },
        "_permissions": ["download"],
        "md5_digest": None,
        "status": "inactive",
        "type": "basic_udm2"
    }

    page_response = {
        "basic_analytic_4b": {
            "_links": {
                "_self": "SELFURL",
                "activate": "ACTIVATEURL",
                "type": "https://api.planet.com/data/v1/asset-types/basic_analytic_4b"
            },
            "_permissions": ["download"],
            "md5_digest": None,
            "status": "inactive",
            "type": "basic_analytic_4b"
        },
        "basic_udm2": basic_udm2_asset
    }

    mock_resp = httpx.Response(HTTPStatus.OK, json=page_response)
    respx.get(assets_url).return_value = mock_resp

    cl = DataClient(session, base_url=TEST_URL)

    with expectation:
        asset = await cl.get_asset(item_type_id, item_id, asset_type_id)
        assert asset == basic_udm2_asset


@respx.mock
@pytest.mark.parametrize("asset_type_id, expectation",
                         [('basic_udm2', does_not_raise()),
                          ('invalid', pytest.raises(exceptions.ClientError))])
def test_get_asset_sync(asset_type_id, expectation, mock_bundles, data_api):
    item_type_id = 'PSScene'
    item_id = '20221003_002705_38_2461'
    assets_url = f'{TEST_URL}/item-types/{item_type_id}/items/{item_id}/assets'
    basic_udm2_asset = {
        "_links": {
            "_self": "SELFURL",
            "activate": "ACTIVATEURL",
            "type": "https://api.planet.com/data/v1/asset-types/basic_udm2"
        },
        "_permissions": ["download"],
        "md5_digest": None,
        "status": "inactive",
        "type": "basic_udm2"
    }

    page_response = {
        "basic_analytic_4b": {
            "_links": {
                "_self": "SELFURL",
                "activate": "ACTIVATEURL",
                "type": "https://api.planet.com/data/v1/asset-types/basic_analytic_4b"
            },
            "_permissions": ["download"],
            "md5_digest": None,
            "status": "inactive",
            "type": "basic_analytic_4b"
        },
        "basic_udm2": basic_udm2_asset
    }

    mock_resp = httpx.Response(HTTPStatus.OK, json=page_response)
    respx.get(assets_url).return_value = mock_resp
    with expectation:
        asset = data_api.get_asset(item_type_id, item_id, asset_type_id)
        assert asset == basic_udm2_asset


@respx.mock
@pytest.mark.anyio
@pytest.mark.parametrize("status, expectation", [('inactive', True),
                                                 ('active', False)])
async def test_activate_asset_success(status, expectation, session):
    activate_url = f'{TEST_URL}/activate'

    mock_resp = httpx.Response(HTTPStatus.OK)
    route = respx.get(activate_url)
    route.return_value = mock_resp

    basic_udm2_asset = {
        "_links": {
            "_self": "SELFURL",
            "activate": activate_url,
            "type": "https://api.planet.com/data/v1/asset-types/basic_udm2"
        },
        "_permissions": ["download"],
        "md5_digest": None,
        "status": status,
        "type": "basic_udm2"
    }

    cl = DataClient(session, base_url=TEST_URL)
    await cl.activate_asset(basic_udm2_asset)

    assert route.called == expectation


@respx.mock
@pytest.mark.parametrize("status, expectation", [('inactive', True),
                                                 ('active', False)])
def test_activate_asset_success_sync(status, expectation, data_api):
    activate_url = f'{TEST_URL}/activate'

    mock_resp = httpx.Response(HTTPStatus.OK)
    route = respx.get(activate_url)
    route.return_value = mock_resp

    basic_udm2_asset = {
        "_links": {
            "_self": "SELFURL",
            "activate": activate_url,
            "type": "https://api.planet.com/data/v1/asset-types/basic_udm2"
        },
        "_permissions": ["download"],
        "md5_digest": None,
        "status": status,
        "type": "basic_udm2"
    }

    data_api.activate_asset(basic_udm2_asset)

    assert route.called == expectation


@respx.mock
@pytest.mark.anyio
async def test_activate_asset_invalid_asset(session):
    cl = DataClient(session, base_url=TEST_URL)

    with pytest.raises(exceptions.ClientError):
        await cl.activate_asset({})


@respx.mock
@pytest.mark.anyio
async def test_wait_asset_success(session):
    asset_url = f'{TEST_URL}/asset'

    basic_udm2_asset = {
        "_links": {
            "_self": asset_url,
            "activate": "ACTIVATEURL",
            "type": "https://api.planet.com/data/v1/asset-types/basic_udm2"
        },
        "_permissions": ["download"],
        "md5_digest": None,
        "status": 'activating',
        "type": "basic_udm2"
    }

    basic_udm2_asset_active = copy.deepcopy(basic_udm2_asset)
    basic_udm2_asset_active['status'] = 'active'

    route = respx.get(asset_url)
    route.side_effect = [
        httpx.Response(HTTPStatus.OK, json=basic_udm2_asset),
        httpx.Response(HTTPStatus.OK, json=basic_udm2_asset),
        httpx.Response(HTTPStatus.OK, json=basic_udm2_asset_active)
    ]

    cl = DataClient(session, base_url=TEST_URL)
    asset = await cl.wait_asset(basic_udm2_asset, delay=0)

    assert asset == basic_udm2_asset_active


@respx.mock
def test_wait_asset_success_sync(data_api):
    asset_url = f'{TEST_URL}/asset'

    basic_udm2_asset = {
        "_links": {
            "_self": asset_url,
            "activate": "ACTIVATEURL",
            "type": "https://api.planet.com/data/v1/asset-types/basic_udm2"
        },
        "_permissions": ["download"],
        "md5_digest": None,
        "status": 'activating',
        "type": "basic_udm2"
    }

    basic_udm2_asset_active = copy.deepcopy(basic_udm2_asset)
    basic_udm2_asset_active['status'] = 'active'

    route = respx.get(asset_url)
    route.side_effect = [
        httpx.Response(HTTPStatus.OK, json=basic_udm2_asset),
        httpx.Response(HTTPStatus.OK, json=basic_udm2_asset),
        httpx.Response(HTTPStatus.OK, json=basic_udm2_asset_active)
    ]

    asset = data_api.wait_asset(basic_udm2_asset, delay=0)

    assert asset == basic_udm2_asset_active


@respx.mock
@pytest.mark.anyio
async def test_wait_asset_max_attempts(session):
    asset_url = f'{TEST_URL}/asset'

    basic_udm2_asset = {
        "_links": {
            "_self": asset_url,
            "activate": "ACTIVATEURL",
            "type": "https://api.planet.com/data/v1/asset-types/basic_udm2"
        },
        "_permissions": ["download"],
        "md5_digest": None,
        "status": 'activating',
        "type": "basic_udm2"
    }

    route = respx.get(asset_url)
    route.side_effect = [
        httpx.Response(HTTPStatus.OK, json=basic_udm2_asset),
        httpx.Response(HTTPStatus.OK, json=basic_udm2_asset),
    ]

    cl = DataClient(session, base_url=TEST_URL)

    with pytest.raises(exceptions.ClientError):
        await cl.wait_asset(basic_udm2_asset, delay=0, max_attempts=1)


@respx.mock
@pytest.mark.anyio
@pytest.mark.parametrize("exists, overwrite",
                         [(False, False), (True, False), (True, True),
                          (False, True)])
async def test_download_asset(exists,
                              overwrite,
                              tmpdir,
                              open_test_img,
                              session):
    # NOTE: this is a slightly edited version of test_download_asset_img from
    # tests/integration/test_orders_api
    dl_url = f'{TEST_URL}/1?token=IAmAToken'

    img_headers = {
        'Content-Type': 'image/tiff',
        'Content-Length': '527',
        'Content-Disposition': 'attachment; filename="img.tif"'
    }

    async def _stream_img():
        data = open_test_img.read()
        v = memoryview(data)

        chunksize = 100
        for i in range(math.ceil(len(v) / (chunksize))):
            yield v[i * chunksize:min((i + 1) * chunksize, len(v))]

    # populate request parameter to avoid respx cloning, which throws
    # an error caused by respx and not this code
    # https://github.com/lundberg/respx/issues/130
    mock_resp = httpx.Response(HTTPStatus.OK,
                               stream=_stream_img(),
                               headers=img_headers,
                               request='donotcloneme')
    respx.get(dl_url).return_value = mock_resp

    basic_udm2_asset = {
        "_links": {
            "_self": "SELFURL",
            "activate": "ACTIVATEURL",
            "type": "https://api.planet.com/data/v1/asset-types/basic_udm2"
        },
        "_permissions": ["download"],
        "md5_digest": None,
        "status": 'active',
        "location": dl_url,
        "type": "basic_udm2"
    }

    cl = DataClient(session, base_url=TEST_URL)

    if exists:
        Path(tmpdir, 'img.tif').write_text('i exist')

    path = await cl.download_asset(basic_udm2_asset,
                                   directory=tmpdir,
                                   overwrite=overwrite)
    assert path.name == 'img.tif'
    assert path.is_file()

    if exists and not overwrite:
        assert path.read_text() == 'i exist'
    else:
        assert len(path.read_bytes()) == 527


@respx.mock
@pytest.mark.anyio
@pytest.mark.parametrize("exists, overwrite",
                         [(False, False), (True, False), (True, True),
                          (False, True)])
async def test_download_asset_sync(exists,
                                   overwrite,
                                   tmpdir,
                                   open_test_img,
                                   data_api):
    # NOTE: this is a slightly edited version of test_download_asset_img from
    # tests/integration/test_orders_api
    dl_url = f'{TEST_URL}/1?token=IAmAToken'

    img_headers = {
        'Content-Type': 'image/tiff',
        'Content-Length': '527',
        'Content-Disposition': 'attachment; filename="img.tif"'
    }

    async def _stream_img():
        data = open_test_img.read()
        v = memoryview(data)

        chunksize = 100
        for i in range(math.ceil(len(v) / (chunksize))):
            yield v[i * chunksize:min((i + 1) * chunksize, len(v))]

    # populate request parameter to avoid respx cloning, which throws
    # an error caused by respx and not this code
    # https://github.com/lundberg/respx/issues/130
    mock_resp = httpx.Response(HTTPStatus.OK,
                               stream=_stream_img(),
                               headers=img_headers,
                               request='donotcloneme')
    respx.get(dl_url).return_value = mock_resp

    basic_udm2_asset = {
        "_links": {
            "_self": "SELFURL",
            "activate": "ACTIVATEURL",
            "type": "https://api.planet.com/data/v1/asset-types/basic_udm2"
        },
        "_permissions": ["download"],
        "md5_digest": None,
        "status": 'active',
        "location": dl_url,
        "type": "basic_udm2"
    }

    if exists:
        Path(tmpdir, 'img.tif').write_text('i exist')

    path = data_api.download_asset(basic_udm2_asset,
                                   directory=tmpdir,
                                   overwrite=overwrite)
    assert path.name == 'img.tif'
    assert path.is_file()

    if exists and not overwrite:
        assert path.read_text() == 'i exist'
    else:
        assert len(path.read_bytes()) == 527


@respx.mock
@pytest.mark.anyio
@pytest.mark.parametrize(
    "hashes_match, md5_entry, expectation",
    [(True, True, does_not_raise()),
     (False, True, pytest.raises(exceptions.ClientError)),
     (True, False, pytest.raises(exceptions.ClientError))])
async def test_validate_checksum(hashes_match, md5_entry, expectation, tmpdir):
    test_bytes = b'foo bar'
    testfile = Path(tmpdir / 'test.txt')
    testfile.write_bytes(test_bytes)

    hash_md5 = hashlib.md5(test_bytes).hexdigest()

    basic_udm2_asset = {
        "_links": {
            "_self": "SELFURL",
            "activate": "ACTIVATEURL",
            "type": "https://api.planet.com/data/v1/asset-types/basic_udm2"
        },
        "_permissions": ["download"],
        "status": 'active',
        "location": "DOWNLOADURL",
        "type": "basic_udm2"
    }

    if md5_entry:
        asset_hash = hash_md5 if hashes_match else 'invalid'
        basic_udm2_asset["md5_digest"] = asset_hash

    with expectation:
        DataClient.validate_checksum(basic_udm2_asset, testfile)


@respx.mock
@pytest.mark.parametrize(
    "hashes_match, md5_entry, expectation",
    [(True, True, does_not_raise()),
     (False, True, pytest.raises(exceptions.ClientError)),
     (True, False, pytest.raises(exceptions.ClientError))])
def test_validate_checksum_sync(hashes_match, md5_entry, expectation, tmpdir):
    test_bytes = b'foo bar'
    testfile = Path(tmpdir / 'test.txt')
    testfile.write_bytes(test_bytes)

    hash_md5 = hashlib.md5(test_bytes).hexdigest()

    basic_udm2_asset = {
        "_links": {
            "_self": "SELFURL",
            "activate": "ACTIVATEURL",
            "type": "https://api.planet.com/data/v1/asset-types/basic_udm2"
        },
        "_permissions": ["download"],
        "status": 'active',
        "location": "DOWNLOADURL",
        "type": "basic_udm2"
    }

    if md5_entry:
        asset_hash = hash_md5 if hashes_match else 'invalid'
        basic_udm2_asset["md5_digest"] = asset_hash

    with expectation:
        DataAPI.validate_checksum(basic_udm2_asset, testfile)



# Copyright 2020 Planet Labs, Inc.
# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Functionality to perform HTTP requests"""
from __future__ import annotations  # https://stackoverflow.com/a/33533514
import asyncio
from collections import Counter
from contextlib import asynccontextmanager
from http import HTTPStatus
import logging
import random
import threading
import time
from typing import AsyncGenerator, Awaitable, Optional, TypeVar

import httpx
from typing_extensions import Literal

from .auth import Auth, AuthType
from . import exceptions, models
from .__version__ import __version__

T = TypeVar("T")

# NOTE: configuration of the session was performed using the data API quick
# search endpoint. These values can be re-tested, tested with a new endpoint or
# refined using session_configuration.py in the scripts directory.

# For how this list was determined, see
# https://github.com/planetlabs/planet-client-python/issues/580
RETRY_EXCEPTIONS = [
    httpx.ConnectError,
    httpx.ReadError,
    httpx.ReadTimeout,
    httpx.RemoteProtocolError,
    exceptions.BadGateway,
    exceptions.TooManyRequests
]
MAX_RETRIES = 5
MAX_RETRY_BACKOFF = 64  # seconds

DEFAULT_READ_TIMEOUT_SECS = 125.0
RATE_LIMIT = 10  # per second
MAX_ACTIVE = 50

LOGGER = logging.getLogger(__name__)


class BaseSession:

    @staticmethod
    def _get_user_agent():
        return 'planet-client-python/' + __version__

    @staticmethod
    def _log_request(request):
        request_pre = f'{request.method} {request.url}'
        LOGGER.info(f'{request_pre} - Sent')
        LOGGER.debug(f'{request_pre} - {request.headers}')
        request.read()
        LOGGER.debug(f'{request_pre} - {request.content}')

    @staticmethod
    def _log_response(response):
        request = response.request
        LOGGER.info(f'{request.method} {request.url} - '
                    f'Status {response.status_code}')
        LOGGER.debug(response.headers)

    @classmethod
    def _convert_and_raise(cls, error):
        response = error.response

        error_types = {
            HTTPStatus.BAD_REQUEST: exceptions.BadQuery,
            HTTPStatus.UNAUTHORIZED: exceptions.InvalidAPIKey,
            HTTPStatus.FORBIDDEN: exceptions.NoPermission,
            HTTPStatus.NOT_FOUND: exceptions.MissingResource,
            HTTPStatus.CONFLICT: exceptions.Conflict,
            HTTPStatus.TOO_MANY_REQUESTS: exceptions.TooManyRequests,
            HTTPStatus.INTERNAL_SERVER_ERROR: exceptions.ServerError,
            HTTPStatus.BAD_GATEWAY: exceptions.BadGateway
        }
        error_type = error_types.get(response.status_code, exceptions.APIError)
        raise error_type(response.text)

    @classmethod
    def _raise_for_status(cls, response):
        if response.is_error:
            try:
                response.raise_for_status()
            except httpx.HTTPStatusError as e:
                e.response.read()
                cls._convert_and_raise(e)

        return


class _Limiter:
    """Limit number of workers and rate of requests.

    Avoids clobbering the API with thousands of async requests.

    Setting rate_limit to zero disables rate (cadence) limiting.
    Setting max_workers to zero disables capping maximum workers.

    This is inspired by asyncio-throttle[1] but altered to enforce cadence
    based on finding that the API returns TooManyRequestError if 2 calls are
    made too close to eachother (even though max rate limit is 5 calls per
    second)[2].

    In investigating options, aiolimiter[3] was also looked at but it seems to
    have odd behavior with httpx [4].

    References:
    [1] https://github.com/hallazzang/asyncio-throttle
    [2] https://github.com/planetlabs/planet-client-python/issues/580#issuecomment-1182752851 # noqa: E501
    [3] https://github.com/mjpieters/aiolimiter
    [4] https://github.com/mjpieters/aiolimiter/issues/73

    The behavior of limiting in communication with live servers can be tested
    and refined using session_configuration.py in the scripts directory.
    """

    def __init__(self, rate_limit=0, max_workers=0):
        # Configuration
        if rate_limit > 0:
            self.cadence = 1.0 / rate_limit
            LOGGER.debug(f'Throttling cadence set to {self.cadence}s.')
        else:
            self.cadence = 0

        self.limit = max_workers
        if self.limit:
            LOGGER.debug(f'Workers capped at {self.limit}.')

        self.retry_interval = 0.01

        # track state
        self._running = 0
        self._last_call = None

    @staticmethod
    def _get_now():
        return time.monotonic()

    async def throttle(self):
        if self.cadence:
            while True:
                now = self._get_now()
                if self._last_call is None:
                    # first call, no need to throttle
                    self._last_call = now
                    break
                elif now - self._last_call >= self.cadence:
                    LOGGER.debug(
                        f'Throught throttle, delta: {now - self._last_call}')
                    self._last_call = now
                    break
                await asyncio.sleep(self.retry_interval)

    async def acquire(self):
        if self.limit:
            while True:
                if self._running < self.limit:
                    self._running += 1
                    LOGGER.debug('Worker acquired.')
                    break
                await asyncio.sleep(self.retry_interval)

    def release(self):
        if self.limit and self._running:
            LOGGER.debug('Worker released.')
            self._running -= 1

    async def __aenter__(self):
        await self.acquire()
        await self.throttle()

    async def __aexit__(self, exc_type, exc, tb):
        self.release()


class Session(BaseSession):
    """Context manager for asynchronous communication with the Planet service.

    The default behavior is to read authentication information stored in the
    secret file. This behavior can be overridden by providing an `auth.Auth`
    instance as an argument.

    Example:
    ```python
    >>> import asyncio
    >>> from planet import Session
    >>>
    >>> async def main():
    ...     async with Session() as sess:
    ...         # communicate with services here
    ...         pass
    ...
    >>> asyncio.run(main())

    ```

    Example:
    ```python
    >>> import async
    >>> from planet import Auth, Session
    >>>
    >>> async def main():
    ...     auth = Auth.from_key('examplekey')
    ...     async with Session(auth=auth) as sess:
    ...         # communicate with services here
    ...         pass
    ...
    >>> asyncio.run(main())

    ```
    """

    def __init__(
        self,
        auth: Optional[AuthType] = None,
        read_timeout_secs: Optional[float] = None,
    ):
        """Initialize a Session.

        Parameters:
            auth: Planet server authentication.
            read_timeout_secs: Maximum time to wait for data to be received.
        """
        if auth is None:
            # Try getting credentials from environment before checking
            # in the secret file, this is the conventional order (AWS
            # CLI, for example.)
            try:
                auth = Auth.from_env()
            except exceptions.PlanetError:
                auth = Auth.from_file()

        if read_timeout_secs is None:
            read_timeout_secs = DEFAULT_READ_TIMEOUT_SECS

        LOGGER.info(
            f'Session read timeout set to {read_timeout_secs} seconds.')
        timeout = httpx.Timeout(10.0, read=read_timeout_secs)

        headers = {
            'User-Agent': self._get_user_agent(), 'X-Planet-App': 'python-sdk'
        }

        self._client = httpx.AsyncClient(auth=auth,
                                         headers=headers,
                                         timeout=timeout,
                                         follow_redirects=True)

        async def alog_request(*args, **kwargs):
            return self._log_request(*args, **kwargs)

        async def alog_response(*args, **kwargs):
            return self._log_response(*args, **kwargs)

        self._client.event_hooks['request'] = [alog_request]
        self._client.event_hooks['response'] = [
            alog_response, self._raise_for_status
        ]

        self.max_retries = MAX_RETRIES
        self.max_retry_backoff = MAX_RETRY_BACKOFF

        self._limiter = _Limiter(rate_limit=RATE_LIMIT, max_workers=MAX_ACTIVE)
        self.outcomes: Counter[str] = Counter()

        # create a dedicated event loop for this httpx session.
        def _start_background_loop(loop):
            asyncio.set_event_loop(loop)
            loop.run_forever()

        self._loop = asyncio.new_event_loop()
        self._loop_thread = threading.Thread(target=_start_background_loop,
                                             args=(self._loop, ),
                                             daemon=True)
        self._loop_thread.start()

    def _call_sync(self, f: Awaitable[T]) -> T:
        return asyncio.run_coroutine_threadsafe(f, self._loop).result()

    @classmethod
    async def _raise_for_status(cls, response):
        if response.is_error:
            try:
                response.raise_for_status()
            except httpx.HTTPStatusError as e:
                await e.response.aread()
                cls._convert_and_raise(e)

        return

    async def __aenter__(self):
        return self

    async def __aexit__(self, *args):
        await self.aclose()

    async def aclose(self):
        await self._client.aclose()

    async def _retry(self, func, *a, **kw):
        """Run an asynchronous request function with retry.

        Retry uses exponential backoff

        Raises:
            planet.exceptions.TooManyRequests: When retry limit is exceeded.
        """
        # NOTE: if we need something more fancy, consider the following libs:
        # * https://pypi.org/project/retry/
        # * https://pypi.org/project/retrying/
        # TODO: consider increasing the scope of retryable exceptions, ex:
        # https://github.com/googleapis/python-storage/blob/1dc6d647b86466bf133
        # fe03ec8d76c541e19c632/google/cloud/storage/retry.py#L23-L30
        # TODO: if throttling is necessary, check out [1] once v1
        # 1. https://github.com/encode/httpx/issues/984
        num_tries = 0
        while True:
            num_tries += 1
            try:
                resp = await func(*a, **kw)
                break
            except Exception as e:
                if type(e) in RETRY_EXCEPTIONS:
                    if num_tries > self.max_retries:
                        raise e
                    else:
                        self.outcomes.update([type(e)])
                        LOGGER.info(f'Try {num_tries}')
                        LOGGER.info(f'Retrying: caught {type(e)}: {e}')
                        wait_time = self._calculate_wait(
                            num_tries, self.max_retry_backoff)
                        LOGGER.info(f'Retrying: sleeping {wait_time}s')
                        await asyncio.sleep(wait_time)
                else:
                    raise e

        self.outcomes.update(['Successful'])
        return resp

    @staticmethod
    def _calculate_wait(num_tries, max_retry_backoff):
        """Calculates retry wait

        Base wait period is calculated as a exponential based on the number of
        tries. Then, a random jitter of up to 999ms is added to the base wait
        to avoid waves of requests in the case of multiple requests. Finally,
        the wait is thresholded to the maximum retry backoff.

        Because threshold is applied after jitter, calculations that hit
        threshold will not have random jitter applied, they will simply result
        in the threshold value being returned.

        Ref:
        * https://developers.planet.com/docs/data/api-mechanics/
        * https://cloud.google.com/iot/docs/how-tos/exponential-backoff
        """
        random_number_milliseconds = random.randint(0, 999) / 1000.0
        calc_wait = 2**num_tries + random_number_milliseconds
        return min(calc_wait, max_retry_backoff)

    async def request(self,
                      method: str,
                      url: str,
                      json: Optional[dict] = None,
                      params: Optional[dict] = None) -> models.Response:
        """Build a request and submit it with retry and limiting.

        Parameters:
            method: HTTP request method.
            url: Location of the API endpoint.
            json: JSON to send.
            params: Values to send in the query string.

        Returns:
            Server response.

        Raises:
            planet.exceptions.APIException: On API error.
            planet.exceptions.ClientError: When retry limit is exceeded.
        """
        if json:
            headers = {'Content-Type': 'application/json'}
        else:
            headers = None

        request = self._client.build_request(method=method,
                                             url=url,
                                             json=json,
                                             params=params,
                                             headers=headers)

        http_response = await self._retry(self._send, request, stream=False)
        return models.Response(http_response)

    async def _send(self, request, stream=False) -> httpx.Response:
        """Send request with with rate/worker limiting."""
        async with self._limiter:
            http_resp = await self._client.send(request, stream=stream)

        return http_resp

    @asynccontextmanager
    async def stream(
        self,
        method: str,
        url: str,
        params: Optional[dict] = None
    ) -> AsyncGenerator[models.StreamingResponse, None]:
        """Submit a request and get the response as a stream context manager.

        Parameters:
            method: HTTP request method.
            url: Location of the API endpoint.

        Returns:
            Context manager providing the streaming response.
        """
        request = self._client.build_request(method=method,
                                             url=url,
                                             params=params)
        http_response = await self._retry(self._send, request, stream=True)
        response = models.StreamingResponse(http_response)
        try:
            yield response
        finally:
            await response.aclose()

    def client(self,
               name: Literal['data', 'orders', 'subscriptions'],
               base_url: Optional[str] = None) -> object:
        """Get a client by its module name.

        Parameters:
            name: one of 'data', 'orders', or 'subscriptions'.

        Returns:
            A client instance.

        Raises:
            ClientError when no such client can be had.

        """
        # To avoid circular dependency.
        from planet.clients import _client_directory

        try:
            return _client_directory[name](self, base_url=base_url)
        except KeyError:
            raise exceptions.ClientError("No such client.")


class AuthSession(BaseSession):
    """Synchronous connection to the Planet Auth service."""

    def __init__(self):
        """Initialize an AuthSession.
        """
        self._client = httpx.Client(timeout=None)
        self._client.headers.update({'User-Agent': self._get_user_agent()})
        self._client.event_hooks['request'] = [self._log_request]
        self._client.event_hooks['response'] = [
            self._log_response, self._raise_for_status
        ]

    def request(self, method: str, url: str, json: dict):
        """Submit a request

        Parameters:
            method: HTTP request method.
            url: Location of the API endpoint.
            json: JSON to send.

        Returns:
            Server response.

        Raises:
            planet.exceptions.APIException: On API error.
        """
        request = self._client.build_request(method=method, url=url, json=json)
        http_resp = self._client.send(request)
        return models.Response(http_resp)

    @classmethod
    def _raise_for_status(cls, response):
        try:
            super()._raise_for_status(response)
        except exceptions.BadQuery:
            raise exceptions.APIError('Not a valid email address.')
        except exceptions.InvalidAPIKey:
            raise exceptions.APIError('Incorrect email or password.')



"""Session module tests."""

import pytest

from planet import DataClient, OrdersClient, SubscriptionsClient, Session
from planet.exceptions import ClientError


@pytest.mark.anyio
@pytest.mark.parametrize("client_name,client_class",
                         [('data', DataClient), ('orders', OrdersClient),
                          ('subscriptions', SubscriptionsClient)])
async def test_session_get_client(client_name, client_class):
    """Get a client from a session."""
    async with Session() as session:
        client = session.client(client_name)
        assert isinstance(client, client_class)


@pytest.mark.anyio
async def test_session_get_client_error():
    """Get an exception when no such client exists."""
    async with Session() as session:
        with pytest.raises(ClientError):
            _ = session.client('bogus')



# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Tests of the Data CLI."""
from http import HTTPStatus
import json
import logging
from pathlib import Path
import math

import httpx
import respx

from click.testing import CliRunner
import pytest

from planet.cli import cli
from planet.clients.data import (LIST_SEARCH_TYPE_DEFAULT,
                                 LIST_SORT_DEFAULT,
                                 SEARCH_SORT_DEFAULT)

LOGGER = logging.getLogger(__name__)

TEST_URL = 'https://api.planet.com/data/v1'
TEST_QUICKSEARCH_URL = f'{TEST_URL}/quick-search'
TEST_SEARCHES_URL = f'{TEST_URL}/searches'
TEST_STATS_URL = f'{TEST_URL}/stats'
VALID_SEARCH_ID = '286469f0b27c476e96c3c4e561f59664'


@pytest.fixture
def invoke():

    def _invoke(extra_args, runner=None, **kwargs):
        runner = runner or CliRunner()
        args = ['data'] + extra_args
        return runner.invoke(cli.main, args=args, **kwargs)

    return _invoke


@pytest.fixture
def item_type():
    return 'PSScene'


@pytest.fixture
def item_id():
    return '20221003_002705_38_2461xx'


@pytest.fixture
def asset_type():
    return 'basic_udm2'


@pytest.fixture
def dl_url():
    return f'{TEST_URL}/1?token=IAmAToken'


@pytest.fixture
def mock_asset_get_response(item_type, item_id, asset_type, dl_url):

    def _func():
        basic_udm2_asset = {
            "_links": {
                "_self": "SELFURL",
                "activate": "ACTIVATEURL",
                "type": "https://api.planet.com/data/v1/asset-types/basic_udm2"
            },
            "_permissions": ["download"],
            "md5_digest": None,
            "status": 'active',
            "location": dl_url,
            "type": asset_type
        }

        page_response = {
            "basic_analytic_4b": {
                "_links": {
                    "_self": "SELFURL",
                    "activate": "ACTIVATEURL",
                    "type": ('https://api.planet.com/data/v1/asset-types/'
                             'basic_analytic_4b')
                },
                "_permissions": ["download"],
                "md5_digest": None,
                "status": "inactive",
                "type": "basic_analytic_4b"
            },
            "basic_udm2": basic_udm2_asset
        }

        # Mock the response for get_asset
        mock_resp_get = httpx.Response(HTTPStatus.OK, json=page_response)
        asset_url = f'{TEST_URL}/item-types/{item_type}/items/{item_id}/assets'
        respx.get(asset_url).return_value = mock_resp_get

    return _func


@pytest.fixture
def item_descriptions(get_test_file_json):
    item_ids = [
        '20220125_075509_67_1061',
        '20220125_075511_17_1061',
        '20220125_075650_17_1061'
    ]
    items = [get_test_file_json(f'data_item_{id}.json') for id in item_ids]
    return items


def test_data_command_registered(invoke):
    """planet-data command prints help and usage message."""
    runner = CliRunner()
    result = invoke(["--help"], runner=runner)
    assert result.exit_code == 0
    assert "Usage" in result.output
    assert "search" in result.output
    assert "search-create" in result.output
    assert "search-get" in result.output
    assert "search-delete" in result.output
    assert "search-update" in result.output
    assert "asset-download" in result.output
    assert "asset-activate" in result.output
    assert "asset-wait" in result.output
    # Add other sub-commands here.


def test_data_search_command_registered(invoke):
    """planet-data search command prints help and usage message."""
    runner = CliRunner()
    result = invoke(["search", "--help"], runner=runner)
    assert result.exit_code == 0
    assert "Usage" in result.output
    assert "limit" in result.output
    assert "name" in result.output
    assert "sort" in result.output
    assert "pretty" in result.output
    assert "help" in result.output
    # Add other sub-commands here.


@pytest.fixture
def search_filter(get_test_file_json):
    filename = 'data_search_filter_2022-01.json'
    return get_test_file_json(filename)


@pytest.fixture
def assert_and_filters_equal():
    """Check for equality when the order of the config list doesn't matter"""

    def _func(filter1, filter2):
        assert filter1.keys() == filter2.keys()
        assert filter1['type'] == filter2['type']

        assert len(filter1['config']) == len(filter2['config'])
        for c in filter1['config']:
            assert c in filter2['config']

    return _func


@respx.mock
def test_data_filter_defaults(invoke, assert_and_filters_equal):

    result = invoke(["filter"])
    assert result.exit_code == 0

    empty_filter = {"type": "AndFilter", "config": []}
    assert_and_filters_equal(json.loads(result.output), empty_filter)


@respx.mock
def test_data_filter_permission(invoke, assert_and_filters_equal):
    result = invoke(["filter", "--permission"])
    assert result.exit_code == 0

    expected_filt = {
        "type": "AndFilter",
        "config": [{
            "type": "PermissionFilter", "config": ["assets:download"]
        }]
    }
    assert_and_filters_equal(json.loads(result.output), expected_filt)


@respx.mock
def test_data_filter_std_quality(invoke, assert_and_filters_equal):
    result = invoke(["filter", '--std-quality'])
    assert result.exit_code == 0

    expected_filt = {
        "type": "AndFilter",
        "config": [{
            "type": "StringInFilter",
            "field_name": "quality_category",
            "config": ["standard"]
        }]
    }
    assert_and_filters_equal(json.loads(result.output), expected_filt)


@respx.mock
@pytest.mark.parametrize("asset, expected",
                         [('ortho_analytic_8b_sr', ['ortho_analytic_8b_sr']),
                          ('ortho_analytic_8b_sr,ortho_analytic_4b_sr',
                           ['ortho_analytic_8b_sr', 'ortho_analytic_4b_sr']),
                          ('ortho_analytic_8b_sr , ortho_analytic_4b_sr',
                           ['ortho_analytic_8b_sr', 'ortho_analytic_4b_sr'])])
def test_data_filter_asset(asset, expected, invoke, assert_and_filters_equal):

    result = invoke(["filter", f'--asset={asset}'])
    assert result.exit_code == 0

    asset_filter = {"type": "AssetFilter", "config": expected}
    expected_filt = {"type": "AndFilter", "config": [asset_filter]}
    assert_and_filters_equal(json.loads(result.output), expected_filt)


@respx.mock
def test_data_filter_date_range_success(invoke, assert_and_filters_equal):
    """Check filter is created correctly and that multiple options results in
    multiple filters"""

    result = invoke(["filter"] + '--date-range field gt 2021-01-01'.split() +
                    '--date-range field2 lt 2022-01-01'.split())
    assert result.exit_code == 0

    date_range_filter1 = {
        "type": "DateRangeFilter",
        "field_name": "field",
        "config": {
            "gt": "2021-01-01T00:00:00Z"
        }
    }
    date_range_filter2 = {
        "type": "DateRangeFilter",
        "field_name": "field2",
        "config": {
            "lt": "2022-01-01T00:00:00Z"
        }
    }

    expected_filt = {
        "type": "AndFilter",
        "config": [date_range_filter1, date_range_filter2]
    }

    assert_and_filters_equal(json.loads(result.output), expected_filt)


@respx.mock
def test_data_filter_date_range_invalid(invoke):
    result = invoke(["filter"] + '--date-range field gt 2021'.split())
    assert result.exit_code == 2


@respx.mock
@pytest.mark.parametrize("geom_fixture",
                         [('geom_geojson'), ('feature_geojson'),
                          ('featurecollection_geojson')])
def test_data_filter_geom(geom_fixture,
                          request,
                          invoke,
                          geom_geojson,
                          assert_and_filters_equal):
    """Ensure that all GeoJSON forms of describing a geometry are handled
    and all result in the same, valid GeometryFilter being created"""

    geom = request.getfixturevalue(geom_fixture)
    geom_str = json.dumps(geom)
    result = invoke(["filter", f'--geom={geom_str}'])
    assert result.exit_code == 0

    geom_filter = {
        "type": "GeometryFilter",
        "field_name": "geometry",
        "config": geom_geojson
    }

    expected_filt = {"type": "AndFilter", "config": [geom_filter]}

    assert_and_filters_equal(json.loads(result.output), expected_filt)


@respx.mock
def test_data_filter_number_in_success(invoke, assert_and_filters_equal):

    result = invoke(["filter"] + '--number-in field 1'.split() +
                    '--number-in field2 2,3.5'.split())
    assert result.exit_code == 0

    number_in_filter1 = {
        "type": "NumberInFilter", "field_name": "field", "config": [1.0]
    }
    number_in_filter2 = {
        "type": "NumberInFilter", "field_name": "field2", "config": [2.0, 3.5]
    }

    expected_filt = {
        "type": "AndFilter", "config": [number_in_filter1, number_in_filter2]
    }

    assert_and_filters_equal(json.loads(result.output), expected_filt)


@respx.mock
def test_data_filter_number_in_badparam(invoke, assert_and_filters_equal):

    result = invoke(["filter"] + '--number-in field 1,str'.split())
    assert result.exit_code == 2


@respx.mock
def test_data_filter_range(invoke, assert_and_filters_equal):
    """Check filter is created correctly, that multiple options results in
    multiple filters, and that floats are processed correctly."""
    result = invoke(["filter"] + '--range field gt 70'.split() +
                    '--range cloud_cover lt 0.5'.split())
    assert result.exit_code == 0

    range_filter1 = {
        "type": "RangeFilter", "field_name": "field", "config": {
            "gt": 70.0
        }
    }
    range_filter2 = {
        "type": "RangeFilter",
        "field_name": "cloud_cover",
        "config": {
            "lt": 0.5
        }
    }

    expected_filt = {
        "type": "AndFilter", "config": [range_filter1, range_filter2]
    }

    assert_and_filters_equal(json.loads(result.output), expected_filt)


@respx.mock
def test_data_filter_string_in(invoke, assert_and_filters_equal):

    result = invoke(["filter"] + '--string-in field foo'.split() +
                    '--string-in field2 foo,bar'.split())
    assert result.exit_code == 0

    string_in_filter1 = {
        "type": "StringInFilter", "field_name": "field", "config": ["foo"]
    }
    string_in_filter2 = {
        "type": "StringInFilter",
        "field_name": "field2",
        "config": ["foo", "bar"]
    }

    expected_filt = {
        "type": "AndFilter", "config": [string_in_filter1, string_in_filter2]
    }

    assert_and_filters_equal(json.loads(result.output), expected_filt)


@respx.mock
def test_data_filter_update(invoke, assert_and_filters_equal):
    """Check filter is created correctly and that multiple options results in
    multiple filters"""

    result = invoke(["filter"] + '--update field gt 2021-01-01'.split() +
                    '--update field2 gte 2022-01-01'.split())
    assert result.exit_code == 0

    update_filter1 = {
        "type": "UpdateFilter",
        "field_name": "field",
        "config": {
            "gt": "2021-01-01T00:00:00Z"
        }
    }
    update_filter2 = {
        "type": "UpdateFilter",
        "field_name": "field2",
        "config": {
            "gte": "2022-01-01T00:00:00Z"
        }
    }

    expected_filt = {
        "type": "AndFilter", "config": [update_filter1, update_filter2]
    }

    assert_and_filters_equal(json.loads(result.output), expected_filt)


@respx.mock
@pytest.mark.parametrize("item_types, expect_success", [('PSScene', True)])
def test_data_search_cmd_item_types(mock_bundles,
                                    item_types,
                                    expect_success,
                                    invoke):
    """Test for planet data search_quick item types, valid and invalid."""
    mock_resp = httpx.Response(HTTPStatus.OK,
                               json={'features': [{
                                   "key": "value"
                               }]})
    respx.post(TEST_QUICKSEARCH_URL).return_value = mock_resp

    result = invoke(['search', item_types])

    if expect_success:
        assert result.exit_code == 0
        assert len(result.output.strip().split('\n')) == 1  # we have 1 feature
    else:
        assert result.exit_code == 2


@respx.mock
@pytest.mark.parametrize("geom_fixture",
                         [('geom_geojson'), ('feature_geojson'),
                          ('featurecollection_geojson'), ('geom_reference'),
                          ("str_geom_reference")])
def test_data_search_cmd_top_level_geom(geom_fixture,
                                        mock_bundles,
                                        request,
                                        invoke):
    """Ensure that all GeoJSON forms of describing a geometry are handled
    and all result in the same, valid GeometryFilter being created"""
    mock_resp = httpx.Response(HTTPStatus.OK,
                               json={'features': [{
                                   "key": "value"
                               }]})
    respx.post(TEST_QUICKSEARCH_URL).return_value = mock_resp
    geom = request.getfixturevalue(geom_fixture)
    if isinstance(geom, dict):
        geom = json.dumps(geom)

    result = invoke(["search", 'PSScene', f"--geom={geom}"])
    assert result.exit_code == 0


@respx.mock
@pytest.mark.parametrize("filter", ['{1:1}', '{"foo"}'])
def test_data_search_cmd_filter_invalid_json(invoke, filter):
    """Test for planet data search_quick.
    Test should fail as filter does not contain valid JSON."""
    mock_resp = httpx.Response(HTTPStatus.OK,
                               json={'features': [{
                                   "key": "value"
                               }]})
    respx.post(TEST_QUICKSEARCH_URL).return_value = mock_resp

    result = invoke(['search', 'PSScene', f'--filter={filter}'])
    assert result.exit_code == 2


@respx.mock
def test_data_search_cmd_filter_success(mock_bundles, invoke):
    """Test for planet data search_quick. Test with multiple item_types.
    Test should succeed as filter contains valid JSON."""
    filter = {
        "type": "DateRangeFilter",
        "field_name": "acquired",
        "config": {
            "gt": "2019-12-31T00:00:00Z", "lte": "2020-01-31T00:00:00Z"
        }
    }

    mock_resp = httpx.Response(HTTPStatus.OK,
                               json={'features': [{
                                   "key": "value"
                               }]})
    respx.post(TEST_QUICKSEARCH_URL).return_value = mock_resp

    result = invoke(['search', 'PSScene', f'--filter={json.dumps(filter)}'])

    assert result.exit_code == 0
    assert len(result.output.strip().split('\n')) == 1  # we have 1 feature

    # check that filter was sent
    expected_request = {"item_types": ['PSScene'], "filter": filter}
    actual_body = json.loads(respx.calls[0].request.content)

    assert actual_body == expected_request


@respx.mock
def test_data_search_cmd_sort_success(mock_bundles, invoke):
    # this cannot be the default value or else the sort param will not be
    # added to the url
    sort = 'published asc'
    search_url = f'{TEST_QUICKSEARCH_URL}?_sort={sort}'

    feature = {"key": "value"}
    mock_resp = httpx.Response(HTTPStatus.OK, json={'features': [feature]})
    respx.post(search_url).return_value = mock_resp

    result = invoke(['search', 'PSScene', f'--sort={sort}'])
    assert result.exit_code == 0
    assert json.loads(result.output) == feature


@respx.mock
def test_data_search_cmd_sort_invalid(invoke):
    result = invoke(['search', 'PSScene', '--sort=invalid'])
    assert result.exit_code == 2


@respx.mock
@pytest.mark.parametrize("limit,limited_list_length", [(None, 100), (0, 102),
                                                       (1, 1)])
def test_data_search_cmd_limit(mock_bundles,
                               invoke,
                               search_results,
                               limit,
                               limited_list_length):
    """Test for planet data search_quick limit option.

    If no value is specified, make sure the result contains at most 100
    entries.
    """
    item_types = 'SkySatCollect'

    # Creating 102 (3x34) search results
    long_search_results = search_results * 34

    all_results = {}
    for x in range(1, len(long_search_results) + 1):
        all_results["result{0}".format(x)] = long_search_results[x - 1]

    page1_response = {
        "_links": {
            "_self": "string1", "assets": "string2", "thumbnail": "string3"
        },
        "features": [
            all_results[f'result{num}']
            for num in range(1, limited_list_length + 1)
        ]
    }
    mock_resp = httpx.Response(HTTPStatus.OK, json=page1_response)

    respx.post(TEST_QUICKSEARCH_URL).return_value = mock_resp

    runner = CliRunner()
    result = invoke(["search", item_types, "--limit", limit], runner=runner)
    assert result.exit_code == 0
    assert result.output.count('"id"') == limited_list_length


@respx.mock
@pytest.mark.parametrize("filter", ['{1:1}', '{"foo"}'])
@pytest.mark.parametrize("item_types",
                         ['PSScene', 'SkySatScene', 'PSScene, SkySatScene'])
def test_data_search_create_filter_invalid_json(invoke, item_types, filter):
    """Test for planet data search_create.
    Test should fail as filter does not contain valid JSON."""
    mock_resp = httpx.Response(HTTPStatus.OK,
                               json={'features': [{
                                   "key": "value"
                               }]})
    respx.post(TEST_SEARCHES_URL).return_value = mock_resp

    name = "temp"

    runner = CliRunner()
    result = invoke([
        "search-create",
        item_types,
        f'--name={name}',
        f'--filter={json.dumps(filter)}'
    ],
                    runner=runner)
    assert result.exit_code == 2


@respx.mock
@pytest.mark.parametrize("item_types",
                         ['PSScene', 'SkySatScene', 'PSScene, SkySatScene'])
def test_data_search_create_filter_success(mock_bundles, invoke, item_types):
    """Test for planet data search_create. Test with multiple item_types.
    Test should succeed as filter contains valid JSON."""
    filter = {
        "type": "DateRangeFilter",
        "field_name": "acquired",
        "config": {
            "gt": "2019-12-31T00:00:00Z", "lte": "2020-01-31T00:00:00Z"
        }
    }

    name = "temp"

    mock_resp = httpx.Response(HTTPStatus.OK,
                               json={'features': [{
                                   "key": "value"
                               }]})
    respx.post(TEST_SEARCHES_URL).return_value = mock_resp

    runner = CliRunner()
    result = invoke([
        "search-create",
        item_types,
        f'--filter={json.dumps(filter)}',
        f'--name={name}'
    ],
                    runner=runner)

    assert result.exit_code == 0
    assert len(result.output.strip().split('\n')) == 1  # we have 1 feature


@respx.mock
def test_data_search_create_daily_email(mock_bundles, invoke, search_result):
    mock_resp = httpx.Response(HTTPStatus.OK, json=search_result)
    respx.post(TEST_SEARCHES_URL).return_value = mock_resp

    filter = {
        "type": "DateRangeFilter",
        "field_name": "acquired",
        "config": {
            "gt": "2019-12-31T00:00:00Z", "lte": "2020-01-31T00:00:00Z"
        }
    }

    result = invoke([
        'search-create',
        'SkySatScene',
        '--name=temp',
        f'--filter={json.dumps(filter)}',
        '--daily-email'
    ])

    search_request = {
        "name": "temp",
        "filter": {
            "type": "DateRangeFilter",
            "field_name": "acquired",
            "config": {
                "gt": "2019-12-31T00:00:00Z", "lte": "2020-01-31T00:00:00Z"
            }
        },
        "item_types": ["SkySatScene"],
        "__daily_email_enabled": True
    }
    sent_request = json.loads(respx.calls.last.request.content)
    assert result.exit_code == 0
    assert sent_request == search_request
    assert json.loads(result.output) == search_result


@respx.mock
@pytest.mark.parametrize("limit, expected_list_length", [(0, 4), (3, 3)])
def test_data_search_list_basic(invoke,
                                search_result,
                                limit,
                                expected_list_length):
    """Ensure planet data search-list runs successfully and respects limit."""
    page1_response = {"_links": {}, "searches": [search_result] * 4}
    route = respx.get(TEST_SEARCHES_URL)
    route.return_value = httpx.Response(200, json=page1_response)

    result = invoke(['search-list', f'--limit={limit}'])
    assert result.exit_code == 0
    assert len(result.output.strip().split('\n')) == expected_list_length


@respx.mock
@pytest.mark.parametrize("sort, rel_url, valid",
                         [(LIST_SORT_DEFAULT, '', True),
                          ('created asc', '?_sort=created+asc', True),
                          ('notvalid', '', False)])
def test_data_search_list_sort(invoke, search_result, sort, rel_url, valid):
    """Ensure planet data search-list handles sort."""
    page1_response = {"_links": {}, "searches": [search_result] * 4}
    route = respx.get(f'{TEST_SEARCHES_URL}{rel_url}')
    route.return_value = httpx.Response(200, json=page1_response)

    result = invoke(['search-list', f'--sort={sort}'])
    expected_code = 0 if valid else 2
    assert result.exit_code == expected_code


@respx.mock
@pytest.mark.parametrize("search_type, rel_url, valid",
                         [(LIST_SEARCH_TYPE_DEFAULT, '', True),
                          ('saved', '?search_type=saved', True),
                          ('notvalid', '', False)])
def test_data_search_list_searchtype(invoke,
                                     search_result,
                                     search_type,
                                     rel_url,
                                     valid):
    """Ensure planet data search-list handles search-type."""
    page1_response = {"_links": {}, "searches": [search_result] * 4}
    route = respx.get(f'{TEST_SEARCHES_URL}{rel_url}')
    route.return_value = httpx.Response(200, json=page1_response)

    result = invoke(['search-list', f'--search-type={search_type}'])
    expected_code = 0 if valid else 2
    assert result.exit_code == expected_code


@respx.mock
@pytest.mark.parametrize("search_id, valid", [(VALID_SEARCH_ID, True),
                                              ('invalid', False)])
@pytest.mark.parametrize("limit, expected_count", [(0, 3), (2, 2)])
def test_data_search_run_basic(invoke,
                               item_descriptions,
                               search_id,
                               valid,
                               limit,
                               expected_count):
    """Ensure planet data search-run runs successfully and handles search id
    and limit."""
    next_page_url = f'{TEST_URL}/blob/?page_marker=IAmATest'
    item1, item2, item3 = item_descriptions
    page1_response = {
        "_links": {
            "_next": next_page_url
        }, "features": [item1, item2]
    }

    route = respx.get(f'{TEST_SEARCHES_URL}/{search_id}/results')
    route.return_value = httpx.Response(204, json=page1_response)

    page2_response = {"_links": {"_self": next_page_url}, "features": [item3]}
    mock_resp2 = httpx.Response(HTTPStatus.OK, json=page2_response)
    respx.get(next_page_url).return_value = mock_resp2

    result = invoke(['search-run', f'--limit={limit}', search_id])

    if valid:
        assert result.exit_code == 0
        assert len(result.output.strip().split('\n')) == expected_count
    else:
        assert result.exit_code == 2


@respx.mock
@pytest.mark.parametrize("sort, rel_url, valid",
                         [(SEARCH_SORT_DEFAULT, '', True),
                          ('acquired asc', '?_sort=acquired+asc', True),
                          ('invalid', '', False)])
def test_data_search_run_sort(invoke, item_descriptions, sort, rel_url, valid):
    """Ensure planet data search-run handles sort."""
    next_page_url = f'{TEST_URL}/blob/?page_marker=IAmATest'
    item1, item2, item3 = item_descriptions
    page1_response = {
        "_links": {
            "_next": next_page_url
        }, "features": [item1, item2]
    }

    route = respx.get(
        f'{TEST_SEARCHES_URL}/{VALID_SEARCH_ID}/results{rel_url}')
    route.return_value = httpx.Response(204, json=page1_response)

    page2_response = {"_links": {"_self": next_page_url}, "features": [item3]}
    mock_resp2 = httpx.Response(HTTPStatus.OK, json=page2_response)
    respx.get(next_page_url).return_value = mock_resp2

    result = invoke(['search-run', f'--sort={sort}', VALID_SEARCH_ID])
    expected_code = 0 if valid else 2
    assert result.exit_code == expected_code


@respx.mock
@pytest.mark.parametrize("filter", ['{1:1}', '{"foo"}'])
def test_data_stats_invalid_filter(invoke, filter):
    """Test for planet data stats. Test with multiple item_types.
    Test should fail as filter does not contain valid JSON."""
    mock_resp = httpx.Response(HTTPStatus.OK,
                               json={'features': [{
                                   "key": "value"
                               }]})
    respx.post(TEST_STATS_URL).return_value = mock_resp
    interval = "hour"
    item_type = 'PSScene'
    runner = CliRunner()
    result = invoke(
        ["stats", item_type, f'--interval={interval}', f'--filter={filter}'],
        runner=runner)
    assert result.exit_code == 2


@respx.mock
@pytest.mark.parametrize("item_types",
                         ['PSScene', 'SkySatScene', 'PSScene, SkySatScene'])
@pytest.mark.parametrize("interval, exit_code", [(None, 2), ('hou', 2),
                                                 ('hour', 0)])
def test_data_stats_interval(invoke,
                             mock_bundles,
                             item_types,
                             interval,
                             exit_code):
    """Test for planet data stats. Test with multiple item_types.
    Test should succeed with valid interval, and fail with invalid interval."""
    filter = {
        "type": "DateRangeFilter",
        "field_name": "acquired",
        "config": {
            "gt": "2019-12-31T00:00:00Z", "lte": "2020-01-31T00:00:00Z"
        }
    }

    mock_resp = httpx.Response(HTTPStatus.OK,
                               json={'features': [{
                                   "key": "value"
                               }]})
    respx.post(TEST_STATS_URL).return_value = mock_resp

    args = ["stats", item_types, f'--filter={json.dumps(filter)}']
    if interval:
        args.append(f'--interval={interval}')

    result = invoke(args)
    assert result.exit_code == exit_code


@respx.mock
@pytest.mark.parametrize("item_types",
                         ['PSScene', 'SkySatScene', 'PSScene, SkySatScene'])
@pytest.mark.parametrize("interval", ['hour', 'day', 'week', 'month', 'year'])
def test_data_stats_success(invoke, mock_bundles, item_types, interval):
    """Test for planet data stats. Test with multiple item_types.

    Test should succeed as filter contains valid JSON, item_types, and
    intervals.
    """
    filter = {
        "type": "DateRangeFilter",
        "field_name": "acquired",
        "config": {
            "gt": "2019-12-31T00:00:00Z", "lte": "2020-01-31T00:00:00Z"
        }
    }

    mock_resp = httpx.Response(HTTPStatus.OK,
                               json={'features': [{
                                   "key": "value"
                               }]})
    respx.post(TEST_STATS_URL).return_value = mock_resp

    result = invoke([
        "stats",
        item_types,
        f'--interval={interval}',
        f'--filter={json.dumps(filter)}'
    ])
    assert result.exit_code == 0


# TODO: basic test for "planet data filter".


@respx.mock
def test_search_get(invoke, search_id, search_result):
    get_url = f'{TEST_SEARCHES_URL}/{search_id}'
    mock_resp = httpx.Response(HTTPStatus.OK, json=search_result)
    respx.get(get_url).return_value = mock_resp

    result = invoke(['search-get', search_id])
    assert not result.exception
    assert search_result == json.loads(result.output)


@respx.mock
def test_search_get_id_not_found(invoke, search_id):
    get_url = f'{TEST_SEARCHES_URL}/{search_id}'
    error_json = {"message": "Error message"}
    mock_resp = httpx.Response(404, json=error_json)
    respx.get(get_url).return_value = mock_resp

    result = invoke(['search-get', search_id])
    assert result.exception
    assert 'Error: {"message":"Error message"}\n' == result.output


@respx.mock
def test_search_delete_success(invoke, search_id, search_result):
    delete_url = f'{TEST_SEARCHES_URL}/{search_id}'
    mock_resp = httpx.Response(HTTPStatus.NO_CONTENT, json=search_result)
    respx.delete(delete_url).return_value = mock_resp

    result = invoke(['search-delete', search_id])

    assert not result.exception


@respx.mock
def test_search_delete_nonexistant_search_id(invoke, search_id, search_result):
    delete_url = f'{TEST_SEARCHES_URL}/{search_id}'
    mock_resp = httpx.Response(404, json=search_result)
    respx.delete(delete_url).return_value = mock_resp

    result = invoke(['search-delete', search_id])

    assert result.exception
    assert result.exit_code == 1


@pytest.mark.parametrize("item_types",
                         ['PSScene', 'SkySatScene', 'PSScene, SkySatScene'])
@respx.mock
def test_search_update_success(invoke,
                               search_id,
                               search_result,
                               item_types,
                               search_filter):
    update_url = f'{TEST_SEARCHES_URL}/{search_id}'
    mock_resp = httpx.Response(HTTPStatus.OK, json=search_result)
    respx.put(update_url).return_value = mock_resp

    name = "search_name"

    result = invoke([
        'search-update',
        search_id,
        item_types,
        f'--filter={json.dumps(search_filter)}',
        f'--name={name}'
    ])

    assert not result.exception


@respx.mock
def test_search_update_fail(invoke, search_id, search_filter):
    update_url = f'{TEST_SEARCHES_URL}/{search_id}'
    error_json = {"message": "Error message"}
    mock_resp = httpx.Response(404, json=error_json)
    respx.put(update_url).return_value = mock_resp

    name = "search_name"
    item_types = "PSScene"

    result = invoke([
        'search-update',
        search_id,
        item_types,
        f'--filter={json.dumps(search_filter)}',
        f'--name={name}'
    ])

    assert result.output.startswith("Error")
    assert result.exception


@respx.mock
@pytest.mark.parametrize("exists, overwrite",
                         [(False, False), (True, False), (True, True),
                          (False, True)])
def test_asset_download_default(invoke,
                                open_test_img,
                                exists,
                                overwrite,
                                mock_asset_get_response,
                                item_type,
                                item_id,
                                asset_type,
                                dl_url):

    mock_asset_get_response()

    img_headers = {
        'Content-Type': 'image/tiff',
        'Content-Length': '527',
        'Content-Disposition': 'attachment; filename="img.tif"'
    }

    async def _stream_img():
        data = open_test_img.read()
        v = memoryview(data)

        chunksize = 100
        for i in range(math.ceil(len(v) / (chunksize))):
            yield v[i * chunksize:min((i + 1) * chunksize, len(v))]

    # Mock the response for download_asset
    mock_resp_download = httpx.Response(HTTPStatus.OK,
                                        stream=_stream_img(),
                                        headers=img_headers,
                                        request='donotcloneme')
    respx.get(dl_url).return_value = mock_resp_download

    runner = CliRunner()
    with runner.isolated_filesystem() as folder:
        if exists:
            Path(folder, 'img.tif').write_bytes(b'01010')

        asset_download_command = [
            'asset-download',
            item_type,
            item_id,
            asset_type,
            f'--directory={Path(folder)}',
            '--filename',
            'img.tif'
        ]
        if overwrite:
            asset_download_command.append('--overwrite')

        result = invoke(asset_download_command, runner=runner)
        assert result.exit_code == 0

        path = Path(folder, 'img.tif')

        assert path.name == 'img.tif'
        assert path.is_file()

        if exists and not overwrite:
            assert len(path.read_bytes()) == 5
            assert len(result.output) == 0
        else:
            assert len(path.read_bytes()) == 527
            assert path.name in result.output


@respx.mock
def test_asset_activate(invoke,
                        mock_asset_get_response,
                        item_type,
                        item_id,
                        asset_type,
                        dl_url):

    mock_asset_get_response()

    # Mock the response for activate_asset
    mock_resp_activate = httpx.Response(HTTPStatus.OK)
    respx.get(dl_url).return_value = mock_resp_activate

    runner = CliRunner()
    result = invoke(['asset-activate', item_type, item_id, asset_type],
                    runner=runner)

    assert not result.exception


@respx.mock
def test_asset_wait(invoke,
                    mock_asset_get_response,
                    item_type,
                    item_id,
                    asset_type,
                    dl_url):

    mock_asset_get_response()

    # Mock the response for wait_asset
    mock_resp_wait = httpx.Response(HTTPStatus.OK)
    respx.get(dl_url).return_value = mock_resp_wait

    runner = CliRunner()
    result = invoke(
        ['asset-wait', item_type, item_id, asset_type, '--delay', '0'],
        runner=runner)

    assert not result.exception
    assert "status: active" in result.output


# @respx.mock
# def test_asset_get(invoke):
#     item_type = 'PSScene'
#     item_id = '20221003_002705_38_2461xx'
#     asset_type_id = 'basic_udm2'
#     dl_url = f'{TEST_URL}/1?token=IAmAToken'

#     basic_udm2_asset = {
#         "_links": {
#             "_self": "SELFURL",
#             "activate": "ACTIVATEURL",
#             "type": "https://api.planet.com/data/v1/asset-types/basic_udm2"
#         },
#         "_permissions": ["download"],
#         "md5_digest": None,
#         "status": 'active',
#         "location": dl_url,
#         "type": "basic_udm2"
#     }

#     page_response = {
#         "basic_analytic_4b": {
#             "_links": {
#                 "_self":
#                 "SELFURL",
#                 "activate":
#                 "ACTIVATEURL",
#                 "type":
#                 "https://api.planet.com/data/v1/asset-types/basic_analytic_4b"
#             },
#             "_permissions": ["download"],
#             "md5_digest": None,
#             "status": "inactive",
#             "type": "basic_analytic_4b"
#         },
#         "basic_udm2": basic_udm2_asset
#     }

#     mock_resp = httpx.Response(HTTPStatus.OK, json=page_response)
#     assets_url = f'{TEST_URL}/item-types/{item_type}/items/{item_id}/assets'
#     respx.get(assets_url).return_value = mock_resp

#     runner = CliRunner()
#     result = invoke(['asset-get', item_type, item_id, asset_type_id],
#                     runner=runner)

#     assert not result.exception
#     assert json.dumps(basic_udm2_asset) in result.output

# TODO: basic test for "planet data search-list".
# TODO: basic test for "planet data search-run".
# TODO: basic test for "planet data item-get".
# TODO: basic test for "planet data stats".


@respx.mock
def test_item_types(invoke, mock_bundles):

    result = invoke(['item-types'])

    expected_item_types = ["SkySatScene", "SkySatCollect", "PSScene"]
    for item_type in expected_item_types:
        assert item_type in result.output
    assert result.exit_code == 0



"""CLI HTTP/auth sessions."""

from planet.auth import Auth
from planet.http import Session


class CliSession(Session):
    """Session with CLI-specific auth and identifying header"""

    def __init__(self):
        super().__init__(Auth.from_file())
        self._client.headers.update({'X-Planet-App': 'python-cli'})



# Copyright 2021 Planet Labs, Inc.
# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Functionality for interacting with GeoJSON and planet references."""
import json
import logging
import typing

import geojson as gj
from jsonschema import Draft7Validator

from .models import Feature
from .constants import DATA_DIR
from .exceptions import GeoJSONError, FeatureError

GEOJSON_TYPES = ["Feature"]

LOGGER = logging.getLogger(__name__)


def split_ref(ref: str) -> typing.Tuple[str, str]:
    """split a feature ref into a tuple of the form (collection_id, feature_id)"""
    validate_ref(ref)

    # after validating the ref, return the collection_id and
    # feature_id parts at the end.
    parts = ref.split("/", 4)
    path = parts[-2:]
    return path[0], path[1]


def as_geom_or_ref(data) -> dict:
    """Extract the geometry from GeoJSON and validate.

    Parameters:
        data: GeoJSON geometry, Feature, or FeatureCollection.

    Returns:
        GeoJSON geometry.

    Raises:
        GeoJSONError: If data is not valid GeoJSON geometry, Feature,
            or FeatureCollection or if more than one Feature is in a
            FeatureCollection.
    """
    if isinstance(data, Feature):
        return as_ref(data.ref)
    if isinstance(data, str):
        return as_ref(data)
    geom_type = data['type']
    if geom_type == 'ref':
        return as_ref(data)
    else:
        geom = geom_from_geojson(data)
        validate_geom_as_geojson(geom)
        return geom


def validate_ref(uri) -> bool:
    if uri is None:
        raise FeatureError("Expected str, not None")
    parts = uri.split("/", 4)
    if parts[0] != "pl:features":
        raise FeatureError("Expected scheme pl:features")
    path = parts[1:]
    if len(path) < 2:
        raise FeatureError("Expected dataset/collection path")
    return True


def convert_ref_to_dict(data: str) -> dict:
    """ Ensure geom reference is in the expected format
        Then convert it into a geometry block

    Parameters:
        data: str, a feature reference
    Returns:
        GeoJSON geometry reference
    """
    if validate_ref(data):
        geom = {
            "type": "ref",
            "content": data,
        }
        return geom
    return dict()


def as_ref(data) -> dict:
    if isinstance(data, str):
        data = convert_ref_to_dict(data)
    if isinstance(data, dict):
        geom_type = data['type']
        if geom_type.lower() != 'ref':
            raise FeatureError(
                f'Invalid geometry reference: {geom_type} is not a reference (the type should be "ref").'
            )
        if "content" not in data:
            raise FeatureError(
                'Invalid geometry reference: Missing content block that contains the reference.'
            )
    return data


def as_polygon(data: dict) -> dict:
    geom = as_geom_or_ref(data)
    geom_type = geom['type']
    if geom_type.lower() != 'polygon':
        raise GeoJSONError(
            f'Invalid geometry type: {geom_type} is not Polygon.')
    return geom


def geom_from_geojson(data: dict) -> dict:
    """Get GeoJSON geometry from GeoJSON.

    Parameters:
        data: GeoJSON geometry, Feature, or FeatureCollection.

    Returns:
        GeoJSON geometry.

    Raises:
        planet.exceptions.GeoJSONError: If data is not valid GeoJSON
        geometry or contains multiple features.
    """
    if set(('coordinates', 'type')).issubset(set(data.keys())):
        # already a geom
        ret = data
    else:
        try:
            # feature
            ret = as_geom_or_ref(data['geometry'])
        except KeyError:
            try:
                # FeatureCollection
                features = data['features']
            except KeyError:
                raise GeoJSONError(f'Invalid GeoJSON: {data}')

            if len(features) > 1:
                raise GeoJSONError(
                    'FeatureCollection has multiple features. Only one feature'
                    ' can be used to get geometry.')

            ret = as_geom_or_ref(features[0])
    return ret


def validate_geom_as_geojson(data: dict):
    """Validate GeoJSON geometry.

    Parameters:
        data: GeoJSON geometry.

    Raises:
        planet.exceptions.GeoJSONError: If data is not a valid GeoJSON
        geometry.
    Returns:
        GeoJSON
    """
    data = geom_from_geojson(data)
    if 'type' not in data:
        raise GeoJSONError('Missing "type" key.')
    if 'coordinates' not in data:
        raise GeoJSONError('Missing "coordinates" key.')

    try:
        cls = getattr(gj, data['type'])
        obj = cls(data['coordinates'])
        if not obj.is_valid:
            raise GeoJSONError(obj.errors())
    except AttributeError as err:
        raise GeoJSONError('Not a GeoJSON geometry type') from err
    except ValueError as err:
        raise GeoJSONError('Not a GeoJSON coordinate value') from err

    return data


def as_featurecollection(features: typing.List[dict]) -> dict:
    """Combine the features in a FeatureCollection.

    Parameters:
        features: GeoJSON Feature objects to combine.

    Raises:
        planet.exceptions.GeoJSONError: If any features are not valid
        GeoJSON Feature objects.
    """

    def _check_all_features(obj):
        for f in obj:
            if not _is_instance_of(f, 'Feature'):
                raise GeoJSONError(f'{f} is not a valid GeoJSON Feature ' +
                                   'object.')

    _check_all_features(features)
    return {'type': 'FeatureCollection', 'features': features}


def _is_instance_of(obj: dict, geojson_type: str) -> bool:
    """Determine if an object is a valid instance of a given GeoJSON type.

    Right now this is only used with the Feature GeoJSON type but this will
    be expanded.

    Parameters:
        obj: Input to check
        geojson_type: Name of geojson type to check


    Returns:
        Indication that the object is a valid instance of specified GeoJSON
        type.

    Raises:
        planet.exceptions.GeoJSONError: If geojson_type does not match a
        supported GeoJSON type.
    """
    try:
        schema_name = next(t + '.json' for t in GEOJSON_TYPES
                           if t.lower() == geojson_type.lower())
    except StopIteration:
        raise GeoJSONError(f'Specified geojson_type ({geojson_type}) does '
                           'not match a supported GeoJSON type.')

    filename = DATA_DIR / schema_name
    with open(filename, 'r') as src:
        schema = json.load(src)

    return Draft7Validator(schema).is_valid(obj)



# Copyright 2022 Planet Labs, PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
import json
import logging

from click.testing import CliRunner

from planet.cli import cli

LOGGER = logging.getLogger(__name__)


def test_cli_collect_stdin_non_features():
    values = [{
        'key11': 'value11', 'key12': 'value12'
    }, {
        'key21': 'value21', 'key22': 'value22'
    }]

    runner = CliRunner()
    sequence = '\n'.join([json.dumps(v) for v in values])
    result = runner.invoke(cli.main, ['collect', '-'], input=sequence)

    assert result.exit_code == 0
    assert json.loads(result.output) == values


def test_cli_collect_stdin_features(feature_geojson):
    feature2 = feature_geojson.copy()
    feature2['properties'] = {'foo': 'bar'}
    values = [feature_geojson, feature2]

    runner = CliRunner()
    sequence = '\n'.join([json.dumps(v) for v in values])
    result = runner.invoke(cli.main, ['collect', '-'], input=sequence)

    assert result.exit_code == 0
    expected = {'type': 'FeatureCollection', 'features': values}
    assert json.loads(result.output) == expected


def test_cli_collect_file(feature_geojson):
    feature2 = feature_geojson.copy()
    feature2['properties'] = {'foo': 'bar'}
    values = [feature_geojson, feature2]

    runner = CliRunner()
    sequence = '\n'.join([json.dumps(v) for v in values])

    with runner.isolated_filesystem():
        with open('input.json', 'w') as f:
            f.write(sequence)

        result = runner.invoke(cli.main, ['collect', 'input.json'])

        assert result.exit_code == 0
        expected = {'type': 'FeatureCollection', 'features': values}
        assert json.loads(result.output) == expected



# Copyright 2021 Planet Labs, Inc.
# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import logging

import pytest

from planet import geojson, exceptions

LOGGER = logging.getLogger(__name__)


@pytest.fixture
def assert_geom_equal():

    def _tuple_to_list(obj):
        return json.loads(json.dumps(obj).replace(")", "]").replace("(", "["))

    def fcn(geom_1, geom_2):
        str_1 = _tuple_to_list(geom_1)
        str_2 = _tuple_to_list(geom_2)
        assert str_1 == str_2

    return fcn


def test_geom_from_geojson_success(geom_geojson,
                                   feature_geojson,
                                   featurecollection_geojson,
                                   assert_geom_equal):
    ggeo = geojson.as_geom_or_ref(geom_geojson)
    assert_geom_equal(ggeo, geom_geojson)

    fgeo = geojson.geom_from_geojson(feature_geojson)
    assert_geom_equal(fgeo, geom_geojson)

    fcgeo = geojson.geom_from_geojson(featurecollection_geojson)
    assert_geom_equal(fcgeo, geom_geojson)


def test_geom_from_geojson_no_geometry(feature_geojson):
    feature_geojson.pop("geometry")
    with pytest.raises(exceptions.GeoJSONError):
        _ = geojson.geom_from_geojson(feature_geojson)


def test_geom_from_geojson_missing_coordinates(geom_geojson):
    geom_geojson.pop("coordinates")
    with pytest.raises(exceptions.GeoJSONError):
        _ = geojson.geom_from_geojson(geom_geojson)


def test_geom_from_geojson_missing_type(geom_geojson):
    geom_geojson.pop("type")
    with pytest.raises(exceptions.GeoJSONError):
        _ = geojson.geom_from_geojson(geom_geojson)


def test_geom_from_geojson_multiple_features(featurecollection_geojson):
    # duplicate the feature
    featurecollection_geojson[
        "features"] = 2 * featurecollection_geojson["features"]
    with pytest.raises(geojson.GeoJSONError):
        _ = geojson.geom_from_geojson(featurecollection_geojson)


def test_validate_geom_as_geojson_invalid_type(geom_geojson):
    geom_geojson["type"] = "invalid"
    with pytest.raises(exceptions.GeoJSONError):
        _ = geojson.validate_geom_as_geojson(geom_geojson)


def test_validate_geom_as_geojson_wrong_type(geom_geojson):
    geom_geojson["type"] = "point"
    with pytest.raises(exceptions.GeoJSONError):
        _ = geojson.validate_geom_as_geojson(geom_geojson)


def test_validate_geom_as_geojson_invalid_coordinates(geom_geojson):
    geom_geojson["coordinates"] = "invalid"
    with pytest.raises(exceptions.GeoJSONError):
        _ = geojson.validate_geom_as_geojson(geom_geojson)


def test_validate_geom_as_geojson_empty_coordinates(geom_geojson):
    geom_geojson["coordinates"] = []
    _ = geojson.validate_geom_as_geojson(geom_geojson)


def test_as_geojson(geom_geojson):
    assert geojson.as_geom_or_ref(geom_geojson) == geom_geojson


def test_as_polygon(geom_geojson):
    assert geojson.as_polygon(geom_geojson) == geom_geojson


def test_as_ref(geom_reference):
    assert geojson.as_ref(geom_reference) == geom_reference


def test_as_str_ref(str_geom_reference):
    geomify_ref = {
        "type": "ref",
        "content": str_geom_reference,
    }
    assert geojson.as_ref(str_geom_reference) == geomify_ref


def test_as_invalid_ref():
    with pytest.raises(exceptions.FeatureError):
        geojson.as_ref("some:nonesense/with/nothing")


def test_as_polygon_wrong_type(point_geom_geojson):
    with pytest.raises(exceptions.GeoJSONError):
        _ = geojson.as_polygon(point_geom_geojson)


def test_as_featurecollection_success(feature_geojson):
    feature2 = feature_geojson.copy()
    feature2["properties"] = {"foo": "bar"}
    values = [feature_geojson, feature2]
    res = geojson.as_featurecollection(values)

    expected = {"type": "FeatureCollection", "features": values}
    assert res == expected


def test__is_instance_of_success(feature_geojson):
    assert geojson._is_instance_of(feature_geojson, "Feature")

    feature2 = feature_geojson.copy()
    feature2["properties"] = {"foo": "bar"}
    assert geojson._is_instance_of(feature2, "Feature")


def test__is_instance_of_does_not_exist(feature_geojson):
    with pytest.raises(exceptions.GeoJSONError):
        geojson._is_instance_of(feature_geojson, "Foobar")



# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""The Planet Data CLI."""
from typing import List, Optional
from contextlib import asynccontextmanager
from pathlib import Path
import click

from planet.reporting import AssetStatusBar
from planet import data_filter, DataClient, exceptions
from planet.clients.data import (SEARCH_SORT,
                                 LIST_SEARCH_TYPE,
                                 LIST_SEARCH_TYPE_DEFAULT,
                                 LIST_SORT_ORDER,
                                 LIST_SORT_DEFAULT,
                                 SEARCH_SORT_DEFAULT,
                                 STATS_INTERVAL)

from planet.specs import (FetchBundlesSpecError,
                          get_item_types,
                          SpecificationException,
                          validate_data_item_type)

from . import types
from .cmds import coro, translate_exceptions
from .io import echo_json
from .options import limit, pretty
from .session import CliSession
from .validators import check_geom


@asynccontextmanager
async def data_client(ctx):
    async with CliSession() as sess:
        cl = DataClient(sess, base_url=ctx.obj['BASE_URL'])
        yield cl


@click.group()  # type: ignore
@click.pass_context
@click.option('-u',
              '--base-url',
              default=None,
              help='Assign custom base Data API URL.')
def data(ctx, base_url):
    """Commands for interacting with the Data API"""
    ctx.obj['BASE_URL'] = base_url


# TODO: filter().
def geom_to_filter(ctx, param, value: Optional[dict]) -> Optional[dict]:
    return data_filter.geometry_filter(value) if value else None


def assets_to_filter(ctx, param, assets: List[str]) -> Optional[dict]:
    # TODO: validate and normalize
    return data_filter.asset_filter(assets) if assets else None


def check_item_types(ctx, param, item_types) -> Optional[List[dict]]:
    """Validates each item types provided by comparing them to all supported
    item types."""
    try:
        for item_type in item_types:
            validate_data_item_type(item_type)
        return item_types
    except SpecificationException as e:
        raise click.BadParameter(str(e))
    except FetchBundlesSpecError as e:
        raise click.ClickException(str(e))


def check_item_type(ctx, param, item_type) -> Optional[List[dict]]:
    """Validates the item type provided by comparing it to all supported
    item types."""
    try:
        validate_data_item_type(item_type)
    except SpecificationException as e:
        raise click.BadParameter(str(e))
    except FetchBundlesSpecError as e:
        raise click.ClickException(str(e))

    return item_type


def check_search_id(ctx, param, search_id) -> str:
    """Ensure search id is a valid hex string"""
    try:
        _ = DataClient._check_search_id(search_id)
    except exceptions.ClientError as e:
        raise click.BadParameter(str(e))
    return search_id


def date_range_to_filter(ctx, param, values) -> Optional[List[dict]]:

    def _func(obj):
        field, comp, value = obj
        kwargs = {'field_name': field, comp: value}
        return data_filter.date_range_filter(**kwargs)

    return [_func(v) for v in values] if values else None


def range_to_filter(ctx, param, values) -> Optional[List[dict]]:

    def _func(obj):
        field, comp, value = obj
        kwargs = {'field_name': field, comp: value}
        return data_filter.range_filter(**kwargs)

    return [_func(v) for v in values] if values else None


def update_to_filter(ctx, param, values) -> Optional[List[dict]]:

    def _func(obj):
        field, comp, value = obj
        kwargs = {'field_name': field, comp: value}
        return data_filter.update_filter(**kwargs)

    return [_func(v) for v in values] if values else None


def number_in_to_filter(ctx, param, values) -> Optional[List[dict]]:

    def _func(obj):
        field, values = obj
        return data_filter.number_in_filter(field_name=field, values=values)

    return [_func(v) for v in values] if values else None


def string_in_to_filter(ctx, param, values) -> Optional[List[dict]]:

    def _func(obj):
        field, values = obj
        return data_filter.string_in_filter(field_name=field, values=values)

    return [_func(v) for v in values] if values else None


@data.command()  # type: ignore
@click.pass_context
@translate_exceptions
@click.option('--asset',
              type=types.CommaSeparatedString(),
              default=None,
              callback=assets_to_filter,
              help="""Filter to items with one or more of specified assets.
    TEXT is a comma-separated list of entries.
    When multiple entries are specified, an implicit 'or' logic is applied.""")
@click.option('--date-range',
              type=click.Tuple(
                  [types.Field(), types.Comparison(), types.DateTime()]),
              callback=date_range_to_filter,
              multiple=True,
              help="""Filter by date range in field.
    FIELD is the name of the field to filter on.
    COMP can be lt, lte, gt, or gte.
    DATETIME can be an RFC3339 or ISO 8601 string.""")
@click.option('--geom',
              type=types.JSON(),
              callback=geom_to_filter,
              help="""Filter to items that overlap a given geometry. Can be a
              json string, filename, or '-' for stdin.""")
@click.option('--number-in',
              type=click.Tuple([types.Field(), types.CommaSeparatedFloat()]),
              multiple=True,
              callback=number_in_to_filter,
              help="""Filter field by numeric in.
    FIELD is the name of the field to filter on.
    VALUE is a comma-separated list of entries.
    When multiple entries are specified, an implicit 'or' logic is applied.""")
@click.option('--permission',
              is_flag=True,
              default=False,
              show_default=True,
              help='Filter to assets with download permissions.')
@click.option('--range',
              'nrange',
              type=click.Tuple([types.Field(), types.Comparison(), float]),
              callback=range_to_filter,
              multiple=True,
              help="""Filter by number range in field.
    FIELD is the name of the field to filter on.
    COMP can be lt, lte, gt, or gte.""")
@click.option('--std-quality',
              is_flag=True,
              default=False,
              show_default=True,
              help='Filter to standard quality.')
@click.option('--string-in',
              type=click.Tuple([types.Field(), types.CommaSeparatedString()]),
              multiple=True,
              callback=string_in_to_filter,
              help="""Filter field by numeric in.
    FIELD is the name of the field to filter on.
    VALUE is a comma-separated list of entries.
    When multiple entries are specified, an implicit 'or' logic is applied.""")
@click.option(
    '--update',
    type=click.Tuple([types.Field(), types.GTComparison(), types.DateTime()]),
    callback=update_to_filter,
    multiple=True,
    help="""Filter to items with changes to a specified field value made after
    a specified date.
    FIELD is the name of the field to filter on.
    COMP can be gt or gte.
    DATETIME can be an RFC3339 or ISO 8601 string.""")
@pretty
def filter(ctx,
           asset,
           date_range,
           geom,
           number_in,
           nrange,
           string_in,
           update,
           permission,
           pretty,
           std_quality):
    """Create a structured item search filter.

    This command provides basic functionality for specifying a filter by
    creating an AndFilter with the filters identified with the options as
    inputs. This is only a subset of the complex filtering supported by the
    API. For advanced filter creation, either create the filter by hand or use
    the Python API.

    If no options are specified, an empty filter is returned which, when used
    in a search, bypasses all search filtering.
    """
    permission = data_filter.permission_filter() if permission else None
    std_quality = data_filter.std_quality_filter() if std_quality else None

    filter_options = (asset,
                      date_range,
                      geom,
                      number_in,
                      nrange,
                      string_in,
                      update,
                      permission,
                      std_quality)

    # options allowing multiples are broken up so one filter is created for
    # each time the option is specified
    # unspecified options are skipped
    filters = []
    for f in filter_options:
        if f:
            if isinstance(f, list):
                filters.extend(f)
            else:
                filters.append(f)

    if filters:
        filt = data_filter.and_filter(filters)
    else:
        # make it explicit that we return an empty filter
        # when no filters are specified
        filt = data_filter.empty_filter()

    echo_json(filt, pretty)


@data.command()  # type: ignore
@click.pass_context
@translate_exceptions
@coro
@click.argument("item_types",
                type=types.CommaSeparatedString(),
                callback=check_item_types)
@click.option("--geom", type=types.Geometry(), callback=check_geom)
@click.option('--filter',
              type=types.JSON(),
              help="""Apply specified filter to search. Can be a json string,
              filename, or '-' for stdin.""")
@limit
@click.option('--name', type=str, help='Name of the saved search.')
@click.option('--sort',
              type=click.Choice(SEARCH_SORT),
              default=SEARCH_SORT_DEFAULT,
              show_default=True,
              help='Field and direction to order results by.')
@pretty
async def search(ctx, item_types, geom, filter, limit, name, sort, pretty):
    """Execute a structured item search.

    This function outputs a series of GeoJSON descriptions, one for each of the
    returned items, optionally pretty-printed.

    ITEM_TYPES is a comma-separated list of item-types to search.

    If --filter is specified, the filter must be JSON and can be a json string,
    filename, or '-' for stdin. If not specified, search results are not
    filtered.

    Quick searches are stored for approximately 30 days and the --name
    parameter will be applied to the stored quick search.
    """
    async with data_client(ctx) as cl:

        async for item in cl.search(item_types,
                                    geometry=geom,
                                    search_filter=filter,
                                    name=name,
                                    sort=sort,
                                    limit=limit):
            echo_json(item, pretty)


@data.command()  # type: ignore
@click.pass_context
@translate_exceptions
@coro
@click.argument("item_types",
                type=types.CommaSeparatedString(),
                callback=check_item_types)
@click.option("--geom", type=types.Geometry(), callback=check_geom)
@click.option(
    '--filter',
    type=types.JSON(),
    required=True,
    help="""Filter to apply to search. Can be a json string, filename,
         or '-' for stdin.""")
@click.option('--name',
              type=str,
              required=True,
              help='Name of the saved search.')
@click.option('--daily-email',
              is_flag=True,
              help='Send a daily email when new results are added.')
@pretty
async def search_create(ctx,
                        item_types,
                        geom,
                        filter,
                        name,
                        daily_email,
                        pretty):
    """Create a new saved structured item search.

    This function outputs a full JSON description of the created search,
    optionally pretty-printed.

    ITEM_TYPES is a comma-separated list of item-types to search.
    """
    async with data_client(ctx) as cl:
        items = await cl.create_search(item_types=item_types,
                                       geometry=geom,
                                       search_filter=filter,
                                       name=name,
                                       enable_email=daily_email)
        echo_json(items, pretty)


@data.command()  # type: ignore
@click.pass_context
@translate_exceptions
@coro
@click.option('--sort',
              type=click.Choice(LIST_SORT_ORDER),
              default=LIST_SORT_DEFAULT,
              show_default=True,
              help='Field and direction to order results by.')
@click.option('--search-type',
              type=click.Choice(LIST_SEARCH_TYPE),
              default=LIST_SEARCH_TYPE_DEFAULT,
              show_default=True,
              help='Search type filter.')
@limit
@pretty
async def search_list(ctx, sort, search_type, limit, pretty):
    """List saved searches.

    This function outputs a full JSON description of the saved searches,
    optionally pretty-printed.
    """
    async with data_client(ctx) as cl:
        async for item in cl.list_searches(sort=sort,
                                           search_type=search_type,
                                           limit=limit):
            echo_json(item, pretty)


@data.command()  # type: ignore
@click.pass_context
@translate_exceptions
@coro
@click.argument('search_id', callback=check_search_id)
@click.option('--sort',
              type=click.Choice(SEARCH_SORT),
              default=SEARCH_SORT_DEFAULT,
              show_default=True,
              help='Field and direction to order results by.')
@limit
@pretty
async def search_run(ctx, search_id, sort, limit, pretty):
    """Execute a saved structured item search.

    This function outputs a series of GeoJSON descriptions, one for each of the
    returned items, optionally pretty-printed.
    """
    async with data_client(ctx) as cl:
        async for item in cl.run_search(search_id, sort=sort, limit=limit):
            echo_json(item, pretty)


@data.command()  # type: ignore
@click.pass_context
@translate_exceptions
@coro
@click.argument("item_types",
                type=types.CommaSeparatedString(),
                callback=check_item_types)
@click.option(
    '--filter',
    type=types.JSON(),
    required=True,
    help="""Filter to apply to search. Can be a json string, filename,
         or '-' for stdin.""")
@click.option('--interval',
              type=click.Choice(STATS_INTERVAL),
              required=True,
              help='The size of the histogram date buckets.')
async def stats(ctx, item_types, filter, interval):
    """Get a bucketed histogram of items matching the filter.

    This function returns a bucketed histogram of results based on the
    item_types, interval, and filter specified.

    """
    async with data_client(ctx) as cl:
        items = await cl.get_stats(item_types=item_types,
                                   search_filter=filter,
                                   interval=interval)
        echo_json(items)


@data.command()  # type: ignore
@click.pass_context
@translate_exceptions
@coro
@pretty
@click.argument('search_id')
async def search_get(ctx, search_id, pretty):
    """Get a saved search.

    This function outputs a full JSON description of the identified saved
    search, optionally pretty-printed.
    """
    async with data_client(ctx) as cl:
        items = await cl.get_search(search_id)
        echo_json(items, pretty)


@data.command()  # type: ignore
@click.pass_context
@translate_exceptions
@coro
@click.argument('search_id')
async def search_delete(ctx, search_id):
    """Delete an existing saved search.
    """
    async with data_client(ctx) as cl:
        await cl.delete_search(search_id)


@data.command()  # type: ignore
@click.pass_context
@translate_exceptions
@coro
@click.argument('search_id')
@click.argument("item_types",
                type=types.CommaSeparatedString(),
                callback=check_item_types)
@click.option(
    '--filter',
    type=types.JSON(),
    required=True,
    help="""Filter to apply to search. Can be a json string, filename,
         or '-' for stdin.""")
@click.option('--name',
              type=str,
              required=True,
              help='Name of the saved search.')
@click.option("--geom",
              type=types.Geometry(),
              callback=check_geom,
              default=None)
@click.option('--daily-email',
              is_flag=True,
              help='Send a daily email when new results are added.')
@pretty
async def search_update(ctx,
                        search_id,
                        item_types,
                        filter,
                        geom,
                        name,
                        daily_email,
                        pretty):
    """Update a saved search with the given search request.

    This function outputs a full JSON description of the updated search,
    optionally pretty-printed.
    """
    async with data_client(ctx) as cl:
        items = await cl.update_search(search_id,
                                       item_types,
                                       search_filter=filter,
                                       name=name,
                                       geometry=geom,
                                       enable_email=daily_email)
        echo_json(items, pretty)


@data.command()  # type: ignore
@click.pass_context
@translate_exceptions
@coro
@click.argument("item_type", type=str, callback=check_item_type)
@click.argument("item_id")
@click.argument("asset_type")
@click.option('--directory',
              default='.',
              help=('Base directory for file download.'),
              type=click.Path(exists=True,
                              resolve_path=True,
                              writable=True,
                              file_okay=False))
@click.option('--filename',
              default=None,
              help=('Custom name to assign to downloaded file.'),
              type=str)
@click.option('--overwrite',
              is_flag=True,
              default=False,
              help=('Overwrite files if they already exist.'))
@click.option('--checksum',
              is_flag=True,
              default=None,
              help=('Verify that checksums match.'))
async def asset_download(ctx,
                         item_type,
                         item_id,
                         asset_type,
                         directory,
                         filename,
                         overwrite,
                         checksum):
    """Download an activated asset.

    This function will fail if the asset state is not activated. Consider
    calling `asset-wait` before this command to ensure the asset is activated.

    If --checksum is provided, the associated checksums given in the manifest
    are compared against the downloaded files to verify that they match.

    If --checksum is provided, files are already downloaded, and --overwrite is
    not specified, this will simply validate the checksums of the files against
    the manifest.

    Output:
    The full path of the downloaded file. If the quiet flag is not set, this
    also provides ANSI download status reporting.
    """
    quiet = ctx.obj['QUIET']
    async with data_client(ctx) as cl:
        asset = await cl.get_asset(item_type, item_id, asset_type)
        path = await cl.download_asset(asset=asset,
                                       filename=filename,
                                       directory=Path(directory),
                                       overwrite=overwrite,
                                       progress_bar=not quiet)
        if checksum:
            cl.validate_checksum(asset, path)


@data.command()  # type: ignore
@click.pass_context
@translate_exceptions
@coro
@click.argument("item_type", type=str, callback=check_item_type)
@click.argument("item_id")
@click.argument("asset_type")
async def asset_activate(ctx, item_type, item_id, asset_type):
    """Activate an asset."""
    async with data_client(ctx) as cl:
        asset = await cl.get_asset(item_type, item_id, asset_type)
        await cl.activate_asset(asset)


@data.command()  # type: ignore
@click.pass_context
@translate_exceptions
@coro
@click.argument("item_type", type=str, callback=check_item_type)
@click.argument("item_id")
@click.argument("asset_type")
@click.option('--delay',
              type=int,
              default=5,
              help='Time (in seconds) between polls.')
@click.option('--max-attempts',
              type=int,
              default=200,
              show_default=True,
              help='Maximum number of polls. Set to zero for no limit.')
async def asset_wait(ctx, item_type, item_id, asset_type, delay, max_attempts):
    """Wait for an asset to be activated.

    Returns when the asset status has reached "activated" and the asset is
    available.
    """
    quiet = ctx.obj['QUIET']
    async with data_client(ctx) as cl:
        asset = await cl.get_asset(item_type, item_id, asset_type)
        with AssetStatusBar(item_type, item_id, asset_type,
                            disable=quiet) as bar:
            status = await cl.wait_asset(asset,
                                         delay,
                                         max_attempts,
                                         callback=bar.update)
    click.echo(status)


# @data.command()
# @click.pass_context
# @translate_exceptions
# @coro
# @click.argument("item_type")
# @click.argument("item_id")
# @click.argument("asset_type_id")
# @pretty
# async def asset_get(ctx, item_type, item_id, asset_type_id, pretty):
#     """Get an item asset."""
#     async with data_client(ctx) as cl:
#         asset = await cl.get_asset(item_type, item_id, asset_type_id)
#     echo_json(asset, pretty)

# TODO: search_run()".
# TODO: item_get()".


@data.command()  # type: ignore
@click.pass_context
@translate_exceptions
def item_types(ctx):
    """Show valid item types."""
    click.echo("Valid item types:")
    for it in get_item_types():
        click.echo(f"- {it}")



# Copyright 2017 Planet Labs, Inc.
# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""CLI main entry point"""
import logging
import sys

import click

import planet

from . import auth, collect, data, orders, subscriptions, features

LOGGER = logging.getLogger(__name__)


@click.group()  # type: ignore
@click.pass_context
@click.option('--quiet',
              is_flag=True,
              default=False,
              help='Disable ANSI control output.')
@click.version_option(version=planet.__version__)
@click.option('--verbosity',
              default="warning",
              help=("Optional: set verbosity level to warning, info, or debug.\
                  Defaults to warning."))
def main(ctx, verbosity, quiet):
    """Planet SDK for Python CLI"""
    _configure_logging(verbosity)

    # ensure that ctx.obj exists and is a dict (in case `cli()` is called
    # by means other than the `if` block below)
    ctx.ensure_object(dict)
    ctx.obj['QUIET'] = quiet


def _configure_logging(verbosity):
    """configure logging via verbosity level, corresponding
    to log levels warning, info and debug respectfully.

    Parameters:
        verbosity -- user input for verbosity.
    Raises:
        click.BadParameter: on unexpected parameter input """
    # make the user input string lowercase & strip leading/trailing spaces
    verbosity_input = verbosity.lower()
    verbosity_input = verbosity_input.strip()

    if verbosity_input == 'warning':
        log_level = logging.WARNING
    elif verbosity_input == 'info':
        log_level = logging.INFO
    elif verbosity_input == 'debug':
        log_level = logging.DEBUG
    else:
        raise click.BadParameter("please set verbosity to \
            warning, info, or debug.")
    logging.basicConfig(
        stream=sys.stderr,
        level=log_level,
        format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')


main.add_command(auth.auth)  # type: ignore
main.add_command(data.data)  # type: ignore
main.add_command(orders.orders)  # type: ignore
main.add_command(subscriptions.subscriptions)  # type: ignore
main.add_command(collect.collect)  # type: ignore
main.add_command(features.features)



# Copyright 2021 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
import logging

import pytest

import click
from click.testing import CliRunner

from unittest import mock
from planet.cli import cli

LOGGER = logging.getLogger(__name__)


@pytest.mark.parametrize("option,quiet", [("", False), ("--quiet", True)])
def test_cli_orders_quiet(option, quiet):
    """Check that --quiet is passed in context to subcommands."""

    # dummy command so we can invoke cli
    @click.command()
    @click.pass_context
    def test(ctx):
        assert ctx.obj['QUIET'] is quiet

    cli.main.add_command(test)

    runner = CliRunner()
    runner.invoke(cli.main, args=[option, 'test'], catch_exceptions=False)


@pytest.mark.parametrize("verbosity,log_level",
                         [("", logging.WARNING),
                          ("--verbosity=info", logging.INFO),
                          ("--verbosity=debug", logging.DEBUG)])
@mock.patch('planet.cli.cli.logging.basicConfig')
def test_cli_info_verbosity(mock_config, verbosity, log_level):
    """Check that main command configures logging with the proper level."""

    def configtest(stream, level, format):
        assert level == log_level

    mock_config.side_effect = configtest

    # dummy command so we can invoke cli
    @click.command()
    def test():
        pass

    cli.main.add_command(test)

    runner = CliRunner()
    runner.invoke(cli.main, args=[verbosity, 'test'], catch_exceptions=False)


def test_cli_invalid_verbosity():
    """Get a BadParameter error for invalid --verbosity."""

    # dummy command so we can invoke cli
    @click.command()
    def test():
        pass

    cli.main.add_command(test)

    runner = CliRunner()
    result = runner.invoke(cli.main,
                           args=["--verbosity=nothing", 'test'],
                           catch_exceptions=False)
    assert result.exit_code == 2
